"""Sample geometric shapes."""
import numpy as np
import pytransform3d.rotations as pr
import pytransform3d.transformations as pt
from scipy.spatial import ConvexHull
from .utils import norm_vector


def randn_point(random_state, scale=1.0):
    """Sample 3D point from standard normal distribution.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    scale : float, optional (default: 1)
        Scale of point sampled from normal distribution.

    Returns
    -------
    point : array, shape (3,)
        3D Point sampled from standard normal distribution.
    """
    return scale * random_state.randn(3)


def randn_direction(random_state):
    """Sample 3D direction from standard normal distribution and normalize it.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    Returns
    -------
    direction : array, shape (3,)
        3D direction: 3D vector of unit length.
    """
    return norm_vector(random_state.randn(3))


def randn_line(random_state, scale=1.0):
    """Sample 3D line.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    scale : float, optional (default: 1)
        Scale of point sampled from normal distribution.

    Returns
    -------
    line_point : array, shape (3,)
        3D Point sampled from normal distribution.

    line_direction : array, shape (3,)
        3D direction: 3D vector of unit length.
    """
    line_point = randn_point(random_state, scale=scale)
    line_direction = randn_direction(random_state)
    return line_point, line_direction


def randn_line_segment(random_state, scale=1.0):
    """Sample 3D line segment.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    scale : float, optional (default: 1)
        Scale of points sampled from normal distribution.

    Returns
    -------

    segment_start : array, shape (3,)
        Start point of segment sampled from a normal distribution.

    segment_end : array, shape (3,)
        End point of segment sampled from a normal distribution.
    """
    return (randn_point(random_state, scale=scale),
            randn_point(random_state, scale=scale))


def randn_plane(random_state, scale=1.0):
    """Sample plane in 3D.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    scale : float, optional (default: 1)
        Scale of point sampled from normal distribution.

    Returns
    -------
    plane_point : array, shape (3,)
        3D Point sampled from normal distribution.

    plane_normal : array, shape (3,)
        Plane normal: 3D vector of unit length.
    """
    plane_point = randn_point(random_state, scale=scale)
    plane_normal = randn_direction(random_state)
    return plane_point, plane_normal


def randn_triangle(random_state):
    """Sample triangle.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    Returns
    -------
    triangle_points : array, shape (3, 3)
        Each row contains a point of the triangle (A, B, C) sampled from a
        standard normal distribution.
    """
    return random_state.randn(3, 3)


def randn_rectangle(random_state, center_scale=1.0, length_scale=1.0):
    """Sample rectangle.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    center_scale : float, optional (default: 1)
        Scale the center point by this factor.

    length_scale : float, optional (default: 1)
        Scale the lengths by this factor.

    Returns
    -------
    rectangle_center : array, shape (3,)
        Center point of the rectangle sampled from a normal distribution with
        standard deviation 'center_scale'.

    rectangle_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal. One direction is
        sampled from a normal distribution. The other one is generated from
        it.

    rectangle_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle sampled from a uniform
        distribution on the interval (0, length_scale].
    """
    rectangle_center = center_scale * randn_point(random_state)
    rectangle_axis1 = randn_direction(random_state)
    rectangle_axis2 = norm_vector(pr.perpendicular_to_vector(rectangle_axis1))
    rectangle_lengths = (1.0 - random_state.rand(2)) * length_scale
    rectangle_axes = np.vstack((rectangle_axis1, rectangle_axis2))
    return rectangle_center, rectangle_axes, rectangle_lengths


def rand_circle(random_state, radius_scale=1.0):
    """Sample circle (or disk).

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    radius_scale : float, optional (default: 1)
        Scaling factor for radius.

    Returns
    -------
    center : array, shape (3,)
        Center of the circle.

    radius : float
        Radius of the circle within (0, radius_scale].

    normal : array, shape (3,)
        Normal to the plane in which the circle lies.
    """
    center = random_state.randn(3)
    radius = (1.0 - random_state.rand()) * radius_scale
    normal = norm_vector(random_state.randn(3))
    return center, radius, normal


def rand_box(random_state, center_scale=1.0, size_scale=1.0):
    """Sample box.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    center_scale : float, optional (default: 1)
        Scaling factor for center.

    size_scale : float, optional (default: 1)
        Scaling factor for size.

    Returns
    -------
    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Sizes of the box along its axes within (0, size_scale].
    """
    box2origin = pt.random_transform(random_state)
    box2origin[:3, 3] *= center_scale
    size = (1.0 - random_state.rand(3)) * size_scale
    return box2origin, size


def rand_capsule(random_state, center_scale=1.0, radius_scale=1.0,
                 height_scale=1.0):
    """Sample capsule.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    center_scale : float, optional (default: 1)
        Scaling factor for center.

    radius_scale : float, optional (default: 1)
        Scaling factor for radius.

    height_scale : float, optional (default: 1)
        Scaling factor for height.

    Returns
    -------
    capsule2origin : array, shape (4, 4)
        Pose of the capsule.

    radius : float
        Radius of the capsule within (0, radius_scale].

    height : float
        Height of the capsule within (0, height_scale].
    """
    capsule2origin = pt.random_transform(random_state)
    capsule2origin[:3, 3] *= center_scale
    radius = (1.0 - random_state.rand()) * radius_scale
    height = (1.0 - random_state.rand()) * height_scale
    return capsule2origin, radius, height


def rand_ellipsoid(random_state, min_radius=0.0):
    """Sample ellipsoid.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    min_radius : float, optional (default: 0)
        Minimum radius of ellipsoid.

    Returns
    -------
    ellipsoid2origin : array, shape (4, 4)
        Pose of the ellipsoid.

    radii : array, shape (3,)
        Radii of the ellipsoid within (min_radius, min_radius + 1].
    """
    ellipsoid2origin = pt.random_transform(random_state)
    radii = min_radius + 1.0 - random_state.rand(3)
    return ellipsoid2origin, radii


def rand_cylinder(random_state, center_scale=1.0, min_radius=0.0,
                  min_length=0.0):
    """Sample cylinder.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    min_radius : float, optional (default: 0)
        Minimum radius of cylinder.

    min_length : float, optional (default: 0)
        Minimum length of cylinder.

    center_scale : float, optional (default: 1)
        Scaling factor for center.

    Returns
    -------
    cylinder2origin : array, shape (4, 4)
        Pose of the cylinder.

    radius : float
        Radius of the cylinder within (min_radius, min_radius + 1].

    length : float
        Length of the cylinder within (min_length, min_length + 1].
    """
    cylinder2origin = pt.random_transform(random_state)
    cylinder2origin[:3, 3] *= center_scale
    radius = min_radius + (1.0 - random_state.rand())
    length = min_length + (1.0 - random_state.rand())
    return cylinder2origin, radius, length


def rand_sphere(random_state, center_scale=1.0, radius_scale=1.0):
    """Sample sphere.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    center_scale : float, optional (default: 1)
        Scaling factor for center.

    radius_scale : float, optional (default: 1)
        Scaling factor for radius.

    Returns
    -------
    center : array, shape (3,)
        Center of the sphere.

    radius : float
        Radius of the sphere within (0, radius_scale].
    """
    center = random_state.randn(3) * center_scale
    radius = (1.0 - random_state.rand()) * radius_scale
    return center, radius


def randn_convex(random_state, n_points=10, center_scale=1.0, std=1.0, return_indices=False):
    """Sample convex mesh.

    Parameters
    ----------
    random_state : np.random.RandomState
        Random number generator.

    n_points : int
        Number of points to sample from normal distribution.

    center_scale : float, optional (default: 1)
        Scaling factor for center.

    std : float, optional (default: 1)
        Standard deviation of normal distribution.

    return_indices : bool, optional (default: False)
        Return indices of vertices in faces and all points.

    Returns
    -------
    vertices : array, shape (n_convex_points, 3)
        Vertices of the convex mesh.

    faces : array, shape (n_triangles, 3, 3)
        Vertices organized as triangles.

    points : array, shape (n_points,), optional
        All points that have been sampled from normal distribution.

    indices : array, shape (n_triangles, 3), optional
        Indices of points forming the simplical facets of the convex hull.
    """
    points = random_state.randn(n_points, 3) * std
    ch = ConvexHull(points, qhull_options="Qf")
    simplices = ch.simplices
    faces = np.array([points[[i, j, k]] for i, j, k in simplices])

    reordered_indices = np.array([2, 1, 0], dtype=int)
    for face_idx in range(len(faces)):
        if pr.angle_between_vectors(
                np.cross(faces[face_idx, 2] - faces[face_idx, 0],
                         faces[face_idx, 1] - faces[face_idx, 0]),
                np.mean(faces[face_idx], axis=0)) < 0.5 * np.pi:
            faces[face_idx] = faces[face_idx, reordered_indices]
            simplices[face_idx] = simplices[face_idx, reordered_indices]

    center = random_state.randn(1, 3) * center_scale
    points += center
    faces += center
    vertices = points[ch.vertices]
    ret = [vertices, faces]
    if return_indices:
        ret.extend([ch.points, simplices])
    return ret



import math
import numpy as np


def point_to_cylinder(point, cylinder2origin, radius, length):
    """Compute the shortest distance between point and cylinder.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    cylinder2origin : array, shape (4, 4)
        Pose of the cylinder.

    radius : float
        Radius of the cylinder.

    length : float
        Length of the cylinder.

    Returns
    -------
    distance : float
        The shortest distance between point and triangle.

    closest_point_cylinder : array, shape (3,)
        Closest point on cylinder.
    """
    # signed distance from point to plane of disk
    diff = point - cylinder2origin[:3, 3]
    dist_to_plane = diff.dot(cylinder2origin[:3, 2])

    # projection of P - C onto plane is Q - C = P - C - dist_to_plane * N
    diff_in_plane = diff - dist_to_plane * cylinder2origin[:3, 2]
    sqr_len = diff_in_plane.dot(diff_in_plane)

    length_in_plane = math.sqrt(sqr_len)
    t = radius
    if length_in_plane != 0.0:
        t /= length_in_plane

    closest_point_cylinder = (
        cylinder2origin[:3, 3] + min(1.0, t) * diff_in_plane
        + np.clip(dist_to_plane, -0.5 * length, 0.5 * length) * cylinder2origin[:3, 2])

    return (np.linalg.norm(point - closest_point_cylinder),
            closest_point_cylinder)



"""Distance calculation of specific geometric shapes.

Note that for distance calculations to triangles, rectangles, boxes, cylinders,
spheres, capsules, and convex meshes GJK is usually the fastest option if
applicable. GJK is not applicable if one of the geometric objects is a line
or a plane and GJK is slower when one of the objects is a point or a line
segment.

In the following table an X indicates that the distance computation between
two geometric objects is implemented.

.. raw:: html

    <table>
      <tr>
        <th></th>
        <th style="transform:rotate(315deg);height: 130px;">point</th>
        <th style="transform:rotate(315deg);height: 130px;">line</th>
        <th style="transform:rotate(315deg);height: 130px;">line segment</th>
        <th style="transform:rotate(315deg);height: 130px;">plane</th>
        <th style="transform:rotate(315deg);height: 130px;">triangle</th>
        <th style="transform:rotate(315deg);height: 130px;">rectangle</th>
        <th style="transform:rotate(315deg);height: 130px;">circle</th>
        <th style="transform:rotate(315deg);height: 130px;">disk</th>
        <th style="transform:rotate(315deg);height: 130px;">box</th>
        <th style="transform:rotate(315deg);height: 130px;">ellipsoid</th>
        <th style="transform:rotate(315deg);height: 130px;">cylinder</th>
      </tr>
      <tr>
        <td>point</td>
        <td>-</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
      </tr>
      <tr>
        <td>line</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>-</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
      </tr>
      <tr>
        <td>line segment</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>-</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
      </tr>
      <tr>
        <td>plane</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>X</td>
        <td>X</td>
        <td>-</td>
      </tr>
      <tr>
        <td>triangle</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
      </tr>
      <tr>
        <td>rectangle</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
      </tr>
      <tr>
        <td>circle</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
      </tr>
      <tr>
        <td>disk</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
      </tr>
      <tr>
        <td>box</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
      </tr>
      <tr>
        <td>ellipsoid</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
      </tr>
      <tr>
        <td>cylinder</td>
        <td>X</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
        <td>-</td>
      </tr>
    </table>
"""
from ._line import (
    point_to_line, line_to_line, point_to_line_segment, line_to_line_segment,
    line_segment_to_line_segment)
from ._plane import (
    point_to_plane, line_to_plane, line_segment_to_plane, plane_to_plane,
    plane_to_triangle, plane_to_rectangle, plane_to_box, plane_to_ellipsoid)
from ._triangle import (
    point_to_triangle, line_to_triangle, line_segment_to_triangle,
    triangle_to_triangle, triangle_to_rectangle)
from ._rectangle import (
    point_to_rectangle, line_to_rectangle, line_segment_to_rectangle,
    rectangle_to_rectangle)
from ._disk import point_to_disk, disk_to_disk
from ._circle import point_to_circle, line_to_circle, line_segment_to_circle
from ._box import (
    point_to_box, line_to_box, line_segment_to_box, rectangle_to_box)
from ._ellipsoid import point_to_ellipsoid
from ._cylinder import point_to_cylinder


__all__ = [
    "point_to_line",
    "point_to_line_segment",
    "point_to_plane",
    "point_to_triangle",
    "point_to_rectangle",
    "point_to_disk",
    "point_to_circle",
    "point_to_box",
    "point_to_ellipsoid",
    "point_to_cylinder",
    "line_to_line",
    "line_to_line_segment",
    "line_to_plane",
    "line_to_triangle",
    "line_to_rectangle",
    "line_to_circle",
    "line_to_box",
    "line_segment_to_line_segment",
    "line_segment_to_plane",
    "line_segment_to_triangle",
    "line_segment_to_rectangle",
    "line_segment_to_circle",
    "line_segment_to_box",
    "plane_to_plane",
    "plane_to_triangle",
    "plane_to_rectangle",
    "plane_to_box",
    "plane_to_ellipsoid",
    "triangle_to_triangle",
    "triangle_to_rectangle",
    "rectangle_to_rectangle",
    "rectangle_to_box",
    "disk_to_disk",
]



"""Utility functions."""
import numpy as np


MAX_FLOAT = np.finfo(float).max


def norm_vector(v):
    """Normalize vector.

    Parameters
    ----------
    v : array, shape (n,)
        nd vector

    Returns
    -------
    u : array, shape (n,)
        nd unit vector with norm 1 or the zero vector
    """
    norm = np.linalg.norm(v)
    if norm == 0.0:
        return v

    return v / norm



"""Gilbert-Johnson-Keerthi (GJK) for distance calculation of convex shapes."""
import math
import numpy as np
from .colliders import Convex


def gjk(vertices1, vertices2):
    """Gilbert-Johnson-Keerthi algorithm for distance calculation.

    Parameters
    ----------
    vertices1 : array, shape (n_vertices1, 3)
        Vertices of the first convex shape.

    vertices2 : array, shape (n_vertices2, 3)
        Vertices of the second convex shape.

    Returns
    -------
    distance : float
        The shortest distance between two convex shapes.

    contact_point1 : array, shape (3,)
        Contact point on first convex shape.

    contact_point2 : array, shape (3,)
        Contact point on second convex shape.
    """
    return gjk_with_simplex(Convex(vertices1), Convex(vertices2))[:3]


def gjk_with_simplex(collider1, collider2):
    """Gilbert-Johnson-Keerthi (GJK) algorithm for distance calculation.

    The GJK algorithm only works for convex shapes. Concave objects have to be
    decomposed into convex shapes first.

    Based on the translation to C of the original Fortran implementation:
    Ruspini, Diego. gilbert.c, a C version of the original Fortran
    implementation of the GJK algorithm.
    ftp://labrea.stanford.edu/cs/robotics/sean/distance/gilbert.c,
    also available from http://realtimecollisiondetection.net/files/gilbert.c

    Parameters
    ----------
    collider1 : ConvexCollider
        Convex collider 1.

    collider2 : ConvexCollider
        Convex collider 2.

    Returns
    -------
    distance : float
        The shortest distance between two convex shapes.

    contact_point1 : array, shape (3,)
        Contact point on first convex shape.

    contact_point2 : array, shape (3,)
        Contact point on second convex shape.

    simplex : array, shape (4, 3)
        Simplex defined by 4 points of the Minkowski difference between
        vertices of the two colliders.
    """
    indices_polytope1 = np.array([0, 0, 0, 0], dtype=int)
    indices_polytope2 = np.array([0, 0, 0, 0], dtype=int)

    barycentric_coordinates = np.zeros(4, dtype=float)
    simplex = np.zeros((4, 3), dtype=float)
    old_simplex = np.zeros((4, 3), dtype=float)
    dot_product_table = np.zeros((4, 4), dtype=float)
    old_dot_product_table = np.zeros((4, 4), dtype=float)
    old_indices_polytope1 = np.zeros(4, dtype=int)
    old_indices_polytope2 = np.zeros(4, dtype=int)
    iord = np.zeros(4, dtype=int)
    search_direction = np.zeros(3, dtype=float)
    backup = 0

    # Initialize simplex to difference of first points of the objects
    ncy = 0
    n_simplex_points = 1
    barycentric_coordinates[0] = 1.0

    simplex[0] = collider1.first_vertex() - collider2.first_vertex()
    dot_product_table[0, 0] = np.dot(simplex[0], simplex[0])

    lastdstsq = dot_product_table[0, 0] + dot_product_table[0, 0] + 1.0
    while True:
        ncy += 1

        # Compute point of minimum norm in the convex hull of the simplex
        dstsq, n_simplex_points, backup = distance_subalgorithm(
            n_simplex_points, indices_polytope1, indices_polytope2, simplex,
            dot_product_table, search_direction, barycentric_coordinates, backup)

        if dstsq >= lastdstsq or n_simplex_points == 4:
            if backup:
                closest_point1 = collider1.compute_point(
                    barycentric_coordinates[:n_simplex_points],
                    indices_polytope1[:n_simplex_points])
                closest_point2 = collider2.compute_point(
                    barycentric_coordinates[:n_simplex_points],
                    indices_polytope2[:n_simplex_points])

                # Make sure intersection has zero distance
                if n_simplex_points == 4:
                    closest_point1[:] = 0.5 * (closest_point1 + closest_point2)
                    closest_point2[:] = closest_point1
                    distance = 0.0
                else:
                    distance = math.sqrt(dstsq)

                return distance, closest_point1, closest_point2, simplex

            backup = 1
            if ncy != 1:
                n_simplex_points = _revert_to_old_simplex(
                    dot_product_table, indices_polytope1, indices_polytope2,
                    old_dot_product_table, old_indices_polytope1,
                    old_indices_polytope2, old_simplex, n_old_simplex_points,
                    simplex)
            continue

        lastdstsq = dstsq

        # Find new supporting point in direction -search_direction:
        # s_(A-B)(-search_direction) = s_A(-search_direction) - s_B(search_direction)
        new_index1, new_vertex1 = collider1.support_function(-search_direction)
        new_index2, new_vertex2 = collider2.support_function(search_direction)
        new_simplex_point = new_vertex1 - new_vertex2

        n_simplex_points = _add_new_point(
            dot_product_table, indices_polytope1, indices_polytope2,
            n_simplex_points, new_index1, new_index2, simplex,
            new_simplex_point)
        n_old_simplex_points = _save_old_simplex(
            dot_product_table, indices_polytope1, indices_polytope2,
            n_simplex_points, old_dot_product_table, old_indices_polytope1,
            old_indices_polytope2, old_simplex, simplex)
        _reorder_simplex(
            dot_product_table, indices_polytope1, indices_polytope2, iord,
            n_simplex_points, old_dot_product_table, old_indices_polytope1,
            old_indices_polytope2, old_simplex, simplex)

    raise RuntimeError("Solution should be found in loop.")


def distance_subalgorithm(
        n_simplex_points, old_indices_polytope1, old_indices_polytope2,
        simplex, dot_product_table, search_direction, barycentric_coordinates,
        backup):
    """Distance subalgorithm.

    Implements, in a very efficient way, the distance subalgorithm
    of finding the near point to the convex hull of four or less points
    in 3D space. The procedure and its efficient FORTRAN implementation
    are both due to D.W. Johnson. Although this subroutine is quite long,
    only a very small part of it will be executed on each call. Refer to
    sections 5 and 6 of the report mentioned in routine DIST3 for details
    concerning the distance subalgorithm. Converted to C be Diego C. Ruspini
    3/25/93.

    This function also determines an affinely independent subset of the
    points such that search_direction near point to the affine hull of the
    points in the subset. The variables n_simplex_points, simplex,
    old_indices_polytope1, old_indices_polytope2 and dot_product_table are
    modified so that, on output, they correspond to this subset of points.

    Parameters
    ----------
    n_simplex_points : int
      The number of points. 1 <= n_simplex_points <= 4.

    simplex : array, shape (n_simplex_points, 3)
      The array containing the points.

    old_indices_polytope1 : array, shape (n_simplex_points,)
        Index vector for first polytope. For k = 1, ..., n_simplex_points,
        simplex[k] = vertices1[old_indices_polytope1[k]]
        - vertices2[old_indices_polytope2[k]].

    old_indices_polytope2 : array, shape (n_simplex_points,)
        Index vectors for first and second polytope. For k = 1, ...,
        n_simplex_points, simplex[k] = vertices1[old_indices_polytope1[k]]
        - vertices2[old_indices_polytope2[k]].

    dot_product_table : array, shape (n_simplex_points, n_simplex_points)
        dot_product_table[i, j] = Inner product of simplex[i] and simplex[j].

    search_direction : array, shape (3,)
        Near point to the convex hull of the points in simplex.

    barycentric_coordinates : array, shape (n_simplex_points,)
        The barycentric coordinates of search_direction, i.e.,
        search_direction = barycentric_coordinates[0]*simplex[1] + ...
        + barycentric_coordinates(n_simplex_points)*simplex[n_simplex_points-1],
        barycentric_coordinates[k] > 0.0 for k=0,...,n_simplex_points-1, and,
        barycentric_coordinates[0] + ...
        + barycentric_coordinates[n_simplex_points-1] = 1.0.

    backup : int
        TODO

    Returns
    -------
    dstsq : float
        Squared distance.

    n_new_simplex_points : int
        The new number of points. 1 <= n_new_simplex_points <= 4.

    backup : int
        TODO
    """
    iord = np.empty(4, dtype=int)
    d1 = np.empty(15, dtype=float)
    d2 = np.empty(15, dtype=float)
    d3 = np.empty(15, dtype=float)
    d4 = np.empty(15, dtype=float)
    zsold = np.empty(3, dtype=float)
    alsd = np.empty(4, dtype=float)

    d1[0] = 1.0
    d2[1] = 1.0
    d3[3] = 1.0
    d4[7] = 1.0

    if not backup:
        # regular distance subalgorithm begins
        if n_simplex_points == 1:
            # case  of  a  single  point ...
            barycentric_coordinates[0] = d1[0]
            search_direction[:] = simplex[0]
            dstsq = dot_product_table[0, 0]
            return dstsq, n_simplex_points, backup
        elif n_simplex_points == 2:
            # case of two points ...
            # check optimality of vertex 1
            d2[2] = dot_product_table[0, 0] - dot_product_table[1, 0]
            if d2[2] <= 0.0:
                n_simplex_points = 1
                barycentric_coordinates[0] = d1[0]
                search_direction[:] = simplex[0]
                dstsq = dot_product_table[0, 0]
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 1-2
            d1[2] = dot_product_table[1, 1] - dot_product_table[1, 0]
            if not (d1[2] <= 0.0 or d2[2] <= 0.0):
                sum = d1[2] + d2[2]
                barycentric_coordinates[0] = d1[2] / sum
                barycentric_coordinates[1] = 1.0 - barycentric_coordinates[0]
                search_direction[:] = simplex[1] + barycentric_coordinates[0] * (simplex[0] - simplex[1])
                dstsq = np.dot(search_direction, search_direction)
                return dstsq, n_simplex_points, backup
            # check optimality of vertex 2
            if d1[2] <= 0.0:
                n_simplex_points = 1
                old_indices_polytope1[0] = old_indices_polytope1[1]
                old_indices_polytope2[0] = old_indices_polytope2[1]
                barycentric_coordinates[0] = d2[1]
                search_direction[:] = simplex[1]
                dstsq = dot_product_table[1, 1]
                simplex[0, :] = simplex[1, :]
                dot_product_table[0, 0] = dot_product_table[1, 1]
                return dstsq, n_simplex_points, backup
        elif n_simplex_points == 3:
            # case of three points ...
            # check optimality of vertex 1
            d2[2] = dot_product_table[0, 0] - dot_product_table[1, 0]
            d3[4] = dot_product_table[0, 0] - dot_product_table[2, 0]
            if not (d2[2] > 0.0 or d3[4] > 0.0):
                n_simplex_points = 1
                barycentric_coordinates[0] = d1[0]
                search_direction[:] = simplex[0]
                dstsq = dot_product_table[0, 0]
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 1-2
            e132 = dot_product_table[1, 0] - dot_product_table[2, 1]
            d1[2] = dot_product_table[1, 1] - dot_product_table[1, 0]
            d3[6] = d1[2] * d3[4] + d2[2] * e132
            if not (d1[2] <= 0.0 or d2[2] <= 0.0 or d3[6] > 0.0):
                n_simplex_points = 2
                sum = d1[2] + d2[2]
                barycentric_coordinates[0] = d1[2] / sum
                barycentric_coordinates[1] = 1.0 - barycentric_coordinates[0]
                search_direction[:] = simplex[1] + barycentric_coordinates[0] * (simplex[0] - simplex[1])
                dstsq = np.dot(search_direction, search_direction)
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 1-3
            e123 = dot_product_table[2, 0] - dot_product_table[2, 1]
            d1[4] = dot_product_table[2, 2] - dot_product_table[2, 0]
            d2[6] = d1[4] * d2[2] + d3[4] * e123
            if not (d1[4] <= 0.0 or d2[6] > 0.0 or d3[4] <= 0.0):
                n_simplex_points = 2
                old_indices_polytope1[1] = old_indices_polytope1[2]
                old_indices_polytope2[1] = old_indices_polytope2[2]
                sum = d1[4] + d3[4]
                barycentric_coordinates[0] = d1[4] / sum
                barycentric_coordinates[1] = 1.0 - barycentric_coordinates[0]
                search_direction[:] = simplex[2] + barycentric_coordinates[0] * (simplex[0] - simplex[2])
                dstsq = np.dot(search_direction, search_direction)
                simplex[1, 0] = simplex[2, 0]
                simplex[1, 1] = simplex[2, 1]
                simplex[1, 2] = simplex[2, 2]
                dot_product_table[1, 0] = dot_product_table[2, 0]
                dot_product_table[1, 1] = dot_product_table[2, 2]
                return dstsq, n_simplex_points, backup
            # check optimality of face 123
            e213 = -e123
            d2[5] = dot_product_table[2, 2] - dot_product_table[2, 1]
            d3[5] = dot_product_table[1, 1] - dot_product_table[2, 1]
            d1[6] = d2[5] * d1[2] + d3[5] * e213
            if not (d1[6] <= 0.0 or d2[6] <= 0.0 or d3[6] <= 0.0):
                sum = d1[6] + d2[6] + d3[6]
                barycentric_coordinates[0] = d1[6] / sum
                barycentric_coordinates[1] = d2[6] / sum
                barycentric_coordinates[2] = 1.0 - barycentric_coordinates[0] - barycentric_coordinates[1]
                search_direction[:] = simplex[2] + barycentric_coordinates[0] * (simplex[0] - simplex[2]) + barycentric_coordinates[1] * (simplex[1] - simplex[2])
                dstsq = np.dot(search_direction, search_direction)
                return dstsq, n_simplex_points, backup
            # check optimality of vertex 2
            if not (d1[2] > 0.0 or d3[5] > 0.0):
                n_simplex_points = 1
                old_indices_polytope1[0] = old_indices_polytope1[1]
                old_indices_polytope2[0] = old_indices_polytope2[1]
                barycentric_coordinates[0] = d2[1]
                search_direction[:] = simplex[1]
                dstsq = dot_product_table[1, 1]
                simplex[0, :] = simplex[1, :]
                dot_product_table[0, 0] = dot_product_table[1, 1]
                return dstsq, n_simplex_points, backup
            # check optimality of vertex 3
            if not (d1[4] > 0.0 or d2[5] > 0.0):
                n_simplex_points = 1
                old_indices_polytope1[0] = old_indices_polytope1[2]
                old_indices_polytope2[0] = old_indices_polytope2[2]
                barycentric_coordinates[0] = d3[3]
                search_direction[:] = simplex[2]
                dstsq = dot_product_table[2, 2]
                simplex[0] = simplex[2]
                dot_product_table[0, 0] = dot_product_table[2, 2]
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 2-3
            if not (d1[6] > 0.0 or d2[5] <= 0.0 or d3[5] <= 0.0):
                n_simplex_points = 2
                old_indices_polytope1[0] = old_indices_polytope1[2]
                old_indices_polytope2[0] = old_indices_polytope2[2]
                sum = d2[5] + d3[5]
                barycentric_coordinates[1] = d2[5] / sum
                barycentric_coordinates[0] = 1.0 - barycentric_coordinates[1]
                search_direction[:] = simplex[2] + barycentric_coordinates[1] * (simplex[1] - simplex[2])
                dstsq = np.dot(search_direction, search_direction)
                simplex[0] = simplex[2]
                dot_product_table[1, 0] = dot_product_table[2, 1]
                dot_product_table[0, 0] = dot_product_table[2, 2]
                return dstsq, n_simplex_points, backup
        elif n_simplex_points == 4:
            # case of four points ...
            # check optimality of vertex 1
            d2[2] = dot_product_table[0, 0] - dot_product_table[1, 0]
            d3[4] = dot_product_table[0, 0] - dot_product_table[2, 0]
            d4[8] = dot_product_table[0, 0] - dot_product_table[3, 0]
            if not (d2[2] > 0.0 or d3[4] > 0.0 or d4[8] > 0.0):
                n_simplex_points = 1
                barycentric_coordinates[0] = d1[0]
                search_direction[:] = simplex[0]
                dstsq = dot_product_table[0, 0]
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 1-2
            e132 = dot_product_table[1, 0] - dot_product_table[2, 1]
            e142 = dot_product_table[1, 0] - dot_product_table[3, 1]
            d1[2] = dot_product_table[1, 1] - dot_product_table[1, 0]
            d3[6] = d1[2] * d3[4] + d2[2] * e132
            d4[11] = d1[2] * d4[8] + d2[2] * e142
            if not (d1[2] <= 0.0 or d2[2] <= 0.0 or d3[6] > 0.0 or d4[11] > 0.0):
                n_simplex_points = 2
                sum = d1[2] + d2[2]
                barycentric_coordinates[0] = d1[2] / sum
                barycentric_coordinates[1] = 1.0 - barycentric_coordinates[0]
                search_direction[:] = simplex[1] + barycentric_coordinates[0] * (simplex[0] - simplex[1])
                dstsq = np.dot(search_direction, search_direction)
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 1-3
            e123 = dot_product_table[2, 0] - dot_product_table[2, 1]
            e143 = dot_product_table[2, 0] - dot_product_table[3, 2]
            d1[4] = dot_product_table[2, 2] - dot_product_table[2, 0]
            d2[6] = d1[4] * d2[2] + d3[4] * e123
            d4[12] = d1[4] * d4[8] + d3[4] * e143
            if not (d1[4] <= 0.0 or d2[6] > 0.0 or d3[4] <= 0.0 or d4[12] > 0.0):
                n_simplex_points = 2
                old_indices_polytope1[1] = old_indices_polytope1[2]
                old_indices_polytope2[1] = old_indices_polytope2[2]
                sum = d1[4] + d3[4]
                barycentric_coordinates[0] = d1[4] / sum
                barycentric_coordinates[1] = 1.0 - barycentric_coordinates[0]
                search_direction[:] = simplex[2] + barycentric_coordinates[0] * (simplex[0] - simplex[2])
                dstsq = np.dot(search_direction, search_direction)
                simplex[1] = simplex[2]
                dot_product_table[1, 0] = dot_product_table[2, 0]
                dot_product_table[1, 1] = dot_product_table[2, 2]
                return dstsq, n_simplex_points, backup
            # check optimality of face 123
            d2[5] = dot_product_table[2, 2] - dot_product_table[2, 1]
            d3[5] = dot_product_table[1, 1] - dot_product_table[2, 1]
            e213 = -e123
            d1[6] = d2[5] * d1[2] + d3[5] * e213
            d4[14] = d1[6] * d4[8] + d2[6] * e142 + d3[6] * e143
            if not (d1[6] <= 0.0 or d2[6] <= 0.0 or d3[6] <= 0.0 or d4[14] > 0.0):
                n_simplex_points = 3
                sum = d1[6] + d2[6] + d3[6]
                barycentric_coordinates[0] = d1[6] / sum
                barycentric_coordinates[1] = d2[6] / sum
                barycentric_coordinates[2] = 1.0 - barycentric_coordinates[0] - barycentric_coordinates[1]
                search_direction[:] = simplex[2] + barycentric_coordinates[0] * (simplex[0] - simplex[2]) + barycentric_coordinates[1] * (simplex[1] - simplex[2])
                dstsq = np.dot(search_direction, search_direction)
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 1-4
            e124 = dot_product_table[3, 0] - dot_product_table[3, 1]
            e134 = dot_product_table[3, 0] - dot_product_table[3, 2]
            d1[8] = dot_product_table[3, 3] - dot_product_table[3, 0]
            d2[11] = d1[8] * d2[2] + d4[8] * e124
            d3[12] = d1[8] * d3[4] + d4[8] * e134
            if not (d1[8] <= 0.0 or d2[11] > 0.0 or d3[12] > 0.0 or d4[8] <= 0.0):
                n_simplex_points = 2
                old_indices_polytope1[1] = old_indices_polytope1[3]
                old_indices_polytope2[1] = old_indices_polytope2[3]
                sum = d1[8] + d4[8]
                barycentric_coordinates[0] = d1[8] / sum
                barycentric_coordinates[1] = 1.0 - barycentric_coordinates[0]
                search_direction[:] = simplex[3] + barycentric_coordinates[0] * (simplex[0] - simplex[3])
                dstsq = np.dot(search_direction, search_direction)
                simplex[1] = simplex[3]
                dot_product_table[1, 0] = dot_product_table[3, 0]
                dot_product_table[1, 1] = dot_product_table[3, 3]
                return dstsq, n_simplex_points, backup
            # check optimality of face 1-2-4
            d2[9] = dot_product_table[3, 3] - dot_product_table[3, 1]
            d4[9] = dot_product_table[1, 1] - dot_product_table[3, 1]
            e214 = -e124
            d1[11] = d2[9] * d1[2] + d4[9] * e214
            d3[14] = d1[11] * d3[4] + d2[11] * e132 + d4[11] * e134
            if not (d1[11] <= 0.0 or d2[11] <= 0.0 or d3[14] > 0.0 or d4[11] <= 0.0):
                n_simplex_points = 3
                old_indices_polytope1[2] = old_indices_polytope1[3]
                old_indices_polytope2[2] = old_indices_polytope2[3]
                sum = d1[11] + d2[11] + d4[11]
                barycentric_coordinates[0] = d1[11] / sum
                barycentric_coordinates[1] = d2[11] / sum
                barycentric_coordinates[2] = 1.0 - barycentric_coordinates[0] - barycentric_coordinates[1]
                search_direction[:] = simplex[3] + barycentric_coordinates[0] * (simplex[0] - simplex[3]) + barycentric_coordinates[1] * (simplex[1] - simplex[3])
                dstsq = np.dot(search_direction, search_direction)
                simplex[2] = simplex[3]
                dot_product_table[2, 0] = dot_product_table[3, 0]
                dot_product_table[2, 1] = dot_product_table[3, 1]
                dot_product_table[2, 2] = dot_product_table[3, 3]
                return dstsq, n_simplex_points, backup
            # check optimality of face 1-3-4
            d3[10] = dot_product_table[3, 3] - dot_product_table[3, 2]
            d4[10] = dot_product_table[2, 2] - dot_product_table[3, 2]
            e314 = -e134
            d1[12] = d3[10] * d1[4] + d4[10] * e314
            d2[14] = d1[12] * d2[2] + d3[12] * e123 + d4[12] * e124
            if not (d1[12] <= 0.0 or d2[14] > 0.0 or d3[12] <= 0.0 or d4[12] <= 0.0):
                n_simplex_points = 3
                old_indices_polytope1[1] = old_indices_polytope1[3]
                old_indices_polytope2[1] = old_indices_polytope2[3]
                sum = d1[12] + d3[12] + d4[12]
                barycentric_coordinates[0] = d1[12] / sum
                barycentric_coordinates[2] = d3[12] / sum
                barycentric_coordinates[1] = 1.0 - barycentric_coordinates[0] - barycentric_coordinates[2]
                search_direction[:] = simplex[3] + barycentric_coordinates[0] * (simplex[0] - simplex[3]) + barycentric_coordinates[2] * (simplex[2] - simplex[3])
                dstsq = np.dot(search_direction, search_direction)
                simplex[1] = simplex[3]
                dot_product_table[1, 0] = dot_product_table[3, 0]
                dot_product_table[1, 1] = dot_product_table[3, 3]
                dot_product_table[2, 1] = dot_product_table[3, 2]
                return dstsq, n_simplex_points, backup
            # check optimality of the hull of all 4 points
            e243 = dot_product_table[2, 1] - dot_product_table[3, 2]
            d4[13] = d2[5] * d4[9] + d3[5] * e243
            e234 = dot_product_table[3, 1] - dot_product_table[3, 2]
            d3[13] = d2[9] * d3[5] + d4[9] * e234
            e324 = -e234
            d2[13] = d3[10] * d2[5] + d4[10] * e324
            d1[14] = d2[13] * d1[2] + d3[13] * e213 + d4[13] * e214
            if not (d1[14] <= 0.0 or d2[14] <= 0.0 or d3[14] <= 0.0 or d4[14] <= 0.0):
                sum = d1[14] + d2[14] + d3[14] + d4[14]
                barycentric_coordinates[0] = d1[14] / sum
                barycentric_coordinates[1] = d2[14] / sum
                barycentric_coordinates[2] = d3[14] / sum
                barycentric_coordinates[3] = 1.0 - barycentric_coordinates[0] - barycentric_coordinates[1] - barycentric_coordinates[2]
                search_direction[:] = barycentric_coordinates[0] * simplex[0] + barycentric_coordinates[1] * simplex[1] + barycentric_coordinates[2] * simplex[2] + barycentric_coordinates[3] * simplex[3]
                dstsq = np.dot(search_direction, search_direction)
                return dstsq, n_simplex_points, backup
            # check optimality of vertex 2
            if not (d1[2] > 0.0 or d3[5] > 0.0 or d4[9] > 0.0):
                n_simplex_points = 1
                old_indices_polytope1[0] = old_indices_polytope1[1]
                old_indices_polytope2[0] = old_indices_polytope2[1]
                barycentric_coordinates[0] = d2[1]
                search_direction[:] = simplex[1]
                dstsq = dot_product_table[1, 1]
                simplex[0] = simplex[1]
                dot_product_table[0, 0] = dot_product_table[1, 1]
                return dstsq, n_simplex_points, backup
            # check optimality of vertex 3
            if not (d1[4] > 0.0 or d2[5] > 0.0 or d4[10] > 0.0):
                n_simplex_points = 1
                old_indices_polytope1[0] = old_indices_polytope1[2]
                old_indices_polytope2[0] = old_indices_polytope2[2]
                barycentric_coordinates[0] = d3[3]
                search_direction[:] = simplex[2]
                dstsq = dot_product_table[2, 2]
                simplex[0] = simplex[2]
                dot_product_table[0, 0] = dot_product_table[2, 2]
                return dstsq, n_simplex_points, backup
            # check optimality of vertex 4
            if not (d1[8] > 0.0 or d2[9] > 0.0 or d3[10] > 0.0):
                n_simplex_points = 1
                old_indices_polytope1[0] = old_indices_polytope1[3]
                old_indices_polytope2[0] = old_indices_polytope2[3]
                barycentric_coordinates[0] = d4[7]
                search_direction[:] = simplex[3]
                dstsq = dot_product_table[3, 3]
                simplex[0] = simplex[3]
                dot_product_table[0, 0] = dot_product_table[3, 3]
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 2-3
            if not (d1[6] > 0.0 or d2[5] <= 0.0 or d3[5] <= 0.0 or d4[13] > 0.0):
                n_simplex_points = 2
                old_indices_polytope1[0] = old_indices_polytope1[2]
                old_indices_polytope2[0] = old_indices_polytope2[2]
                sum = d2[5] + d3[5]
                barycentric_coordinates[1] = d2[5] / sum
                barycentric_coordinates[0] = 1.0 - barycentric_coordinates[1]
                search_direction[:] = simplex[2] + barycentric_coordinates[1] * (simplex[1] - simplex[2])
                dstsq = np.dot(search_direction, search_direction)
                simplex[0] = simplex[2]
                dot_product_table[1, 0] = dot_product_table[2, 1]
                dot_product_table[0, 0] = dot_product_table[2, 2]
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 2-4
            if not (d1[11] > 0.0 or d2[9] <= 0.0 or d3[13] > 0.0 or d4[9] <= 0.0):
                n_simplex_points = 2
                old_indices_polytope1[0] = old_indices_polytope1[3]
                old_indices_polytope2[0] = old_indices_polytope2[3]
                sum = d2[9] + d4[9]
                barycentric_coordinates[1] = d2[9] / sum
                barycentric_coordinates[0] = 1.0 - barycentric_coordinates[1]
                search_direction[:] = simplex[3] + barycentric_coordinates[1] * (simplex[1] - simplex[3])
                dstsq = np.dot(search_direction, search_direction)
                simplex[0] = simplex[3]
                dot_product_table[1, 0] = dot_product_table[3, 1]
                dot_product_table[0, 0] = dot_product_table[3, 3]
                return dstsq, n_simplex_points, backup
            # check optimality of line segment 3-4
            if not (d1[12] > 0.0 or d2[13] > 0.0 or d3[10] <= 0.0 or d4[10] <= 0.0):
                n_simplex_points = 2
                old_indices_polytope1[0] = old_indices_polytope1[2]
                old_indices_polytope1[1] = old_indices_polytope1[3]
                old_indices_polytope2[0] = old_indices_polytope2[2]
                old_indices_polytope2[1] = old_indices_polytope2[3]
                sum = d3[10] + d4[10]
                barycentric_coordinates[0] = d3[10] / sum
                barycentric_coordinates[1] = 1.0 - barycentric_coordinates[0]
                search_direction[:] = simplex[3] + barycentric_coordinates[0] * (simplex[2] - simplex[3])
                dstsq = np.dot(search_direction, search_direction)
                simplex[0] = simplex[2]
                simplex[1] = simplex[3]
                dot_product_table[0, 0] = dot_product_table[2, 2]
                dot_product_table[1, 0] = dot_product_table[3, 2]
                dot_product_table[1, 1] = dot_product_table[3, 3]
                return dstsq, n_simplex_points, backup
            # check optimality of face 2-3-4
            if not (d1[14] > 0.0 or d2[13] <= 0.0 or d3[13] <= 0.0 or d4[13] <= 0.0):
                n_simplex_points = 3
                old_indices_polytope1[0] = old_indices_polytope1[3]
                old_indices_polytope2[0] = old_indices_polytope2[3]
                sum = d2[13] + d3[13] + d4[13]
                barycentric_coordinates[1] = d2[13] / sum
                barycentric_coordinates[2] = d3[13] / sum
                barycentric_coordinates[0] = 1.0 - barycentric_coordinates[1] - barycentric_coordinates[2]
                search_direction[:] = simplex[3] + barycentric_coordinates[1] * (simplex[1] - simplex[3]) + barycentric_coordinates[2] * (simplex[2] - simplex[3])
                dstsq = np.dot(search_direction, search_direction)
                simplex[0] = simplex[3]
                dot_product_table[0, 0] = dot_product_table[3, 3]
                dot_product_table[1, 0] = dot_product_table[3, 1]
                dot_product_table[2, 0] = dot_product_table[3, 2]
                return dstsq, n_simplex_points, backup
        else:
            raise ValueError("Invalid value for nvs %d given" % n_simplex_points)

    # ======================================================================
    # The backup procedure  begins ...
    # ======================================================================
    if n_simplex_points == 1:
        # case of a single point ...
        dstsq = dot_product_table[0, 0]
        barycentric_coordinates[0] = d1[0]
        search_direction[:] = simplex[0]
        backup = 1
        return dstsq, n_simplex_points, backup
    elif n_simplex_points == 2:
        # case of two points ...
        if backup:
            d2[2] = dot_product_table[0, 0] - dot_product_table[1, 0]
            d1[2] = dot_product_table[1, 1] - dot_product_table[1, 0]
        # check vertex 1
        dstsq = dot_product_table[0, 0]
        nvsd = 1
        barycentric_coordinates[0] = d1[0]
        search_direction[:] = simplex[0]
        iord[0] = 0
        # check line segment 1-2
        if not (d1[2] <= 0.0 or d2[2] <= 0.0):
            sum = d1[2] + d2[2]
            alsd[0] = d1[2] / sum
            alsd[1] = 1.0 - alsd[0]
            zsold[:] = simplex[1, :] + alsd[0] * (simplex[0, :] - simplex[1, :])
            dstsqd = zsold[0] * zsold[0] + zsold[1] * zsold[1] + zsold[2] * zsold[2]
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold[:]
                iord[0] = 0
                iord[1] = 1
        # check vertex 2
        if dot_product_table[1, 1] < dstsq:
            dstsq = dot_product_table[1, 1]
            nvsd = 1
            barycentric_coordinates[0] = d2[1]
            search_direction[:] = simplex[1]
            iord[0] = 1
    elif n_simplex_points == 3:
        # case of three points
        if backup:
            d2[2] = dot_product_table[0, 0] - dot_product_table[1, 0]
            d3[4] = dot_product_table[0, 0] - dot_product_table[2, 0]
            e132 = dot_product_table[1, 0] - dot_product_table[2, 1]
            d1[2] = dot_product_table[1, 1] - dot_product_table[1, 0]
            d3[6] = d1[2] * d3[4] + d2[2] * e132
            e123 = dot_product_table[2, 0] - dot_product_table[2, 1]
            d1[4] = dot_product_table[2, 2] - dot_product_table[2, 0]
            d2[6] = d1[4] * d2[2] + d3[4] * e123
            e213 = -e123
            d2[5] = dot_product_table[2, 2] - dot_product_table[2, 1]
            d3[5] = dot_product_table[1, 1] - dot_product_table[2, 1]
            d1[6] = d2[5] * d1[2] + d3[5] * e213
        # check vertex 1
        dstsq = dot_product_table[0, 0]
        nvsd = 1
        barycentric_coordinates[0] = d1[0]
        search_direction[:] = simplex[0]
        iord[0] = 0
        # check line segment 1-2
        if not (d1[2] <= 0.0 or d2[2] <= 0.0):
            sum = d1[2] + d2[2]
            alsd[0] = d1[2] / sum
            alsd[1] = 1.0 - alsd[0]
            zsold[:] = simplex[1] + alsd[0] * (simplex[0] - simplex[1])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 1
        # check line segment 1-3
        if not (d1[4] <= 0.0 or d3[4] <= 0.0):
            sum = d1[4] + d3[4]
            alsd[0] = d1[4] / sum
            alsd[1] = 1.0 - alsd[0]
            zsold[:] = simplex[2] + alsd[0] * (simplex[0] - simplex[2])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 2
        # check face 1-2-3
        if not (d1[6] <= 0.0 or d2[6] <= 0.0 or d3[6] <= 0.0):
            sum = d1[6] + d2[6] + d3[6]
            alsd[0] = d1[6] / sum
            alsd[1] = d2[6] / sum
            alsd[2] = 1.0 - alsd[0] - alsd[1]
            zsold[:] = simplex[2] + alsd[0] * (simplex[0] - simplex[2]) + alsd[1] * (simplex[1] - simplex[2])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 3
                barycentric_coordinates[:] = alsd
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 1
                iord[2] = 2
        # check vertex 2
        if dot_product_table[1, 1] < dstsq:
            nvsd = 1
            dstsq = dot_product_table[1, 1]
            barycentric_coordinates[0] = d2[1]
            search_direction[:] = simplex[1]
            iord[0] = 1
        # check vertex 3
        if dot_product_table[2, 2] < dstsq:
            nvsd = 1
            dstsq = dot_product_table[2, 2]
            barycentric_coordinates[0] = d3[3]
            search_direction[:] = simplex[2]
            iord[0] = 2
        # check line segment 2-3
        if not (d2[5] <= 0.0 or d3[5] <= 0.0):
            sum = d2[5] + d3[5]
            alsd[1] = d2[5] / sum
            alsd[0] = 1.0 - alsd[1]
            zsold[:] = simplex[2] + alsd[1] * (simplex[1] - simplex[2])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 2
                iord[1] = 1
    elif n_simplex_points == 4:
        # case of four points
        if backup:
            d2[2] = dot_product_table[0, 0] - dot_product_table[1, 0]
            d3[4] = dot_product_table[0, 0] - dot_product_table[2, 0]
            d4[8] = dot_product_table[0, 0] - dot_product_table[3, 0]
            e132 = dot_product_table[1, 0] - dot_product_table[2, 1]
            e142 = dot_product_table[1, 0] - dot_product_table[3, 1]
            d1[2] = dot_product_table[1, 1] - dot_product_table[1, 0]
            d3[6] = d1[2] * d3[4] + d2[2] * e132
            d4[11] = d1[2] * d4[8] + d2[2] * e142
            e123 = dot_product_table[2, 0] - dot_product_table[2, 1]
            e143 = dot_product_table[2, 0] - dot_product_table[3, 2]
            d1[4] = dot_product_table[2, 2] - dot_product_table[2, 0]
            d2[6] = d1[4] * d2[2] + d3[4] * e123
            d4[12] = d1[4] * d4[8] + d3[4] * e143
            d2[5] = dot_product_table[2, 2] - dot_product_table[2, 1]
            d3[5] = dot_product_table[1, 1] - dot_product_table[2, 1]
            e213 = -e123
            d1[6] = d2[5] * d1[2] + d3[5] * e213
            d4[14] = d1[6] * d4[8] + d2[6] * e142 + d3[6] * e143
            e124 = dot_product_table[3, 0] - dot_product_table[3, 1]
            e134 = dot_product_table[3, 0] - dot_product_table[3, 2]
            d1[8] = dot_product_table[3, 3] - dot_product_table[3, 0]
            d2[11] = d1[8] * d2[2] + d4[8] * e124
            d3[12] = d1[8] * d3[4] + d4[8] * e134
            d2[9] = dot_product_table[3, 3] - dot_product_table[3, 1]
            d4[9] = dot_product_table[1, 1] - dot_product_table[3, 1]
            e214 = -e124
            d1[11] = d2[9] * d1[2] + d4[9] * e214
            d3[14] = d1[11] * d3[4] + d2[11] * e132 + d4[11] * e134
            d3[10] = dot_product_table[3, 3] - dot_product_table[3, 2]
            d4[10] = dot_product_table[2, 2] - dot_product_table[3, 2]
            e314 = -e134
            d1[12] = d3[10] * d1[4] + d4[10] * e314
            d2[14] = d1[12] * d2[2] + d3[12] * e123 + d4[12] * e124
            e243 = dot_product_table[2, 1] - dot_product_table[3, 2]
            d4[13] = d2[5] * d4[9] + d3[5] * e243
            e234 = dot_product_table[3, 1] - dot_product_table[3, 2]
            d3[13] = d2[9] * d3[5] + d4[9] * e234
            e324 = -e234
            d2[13] = d3[10] * d2[5] + d4[10] * e324
            d1[14] = d2[13] * d1[2] + d3[13] * e213 + d4[13] * e214
        # check vertex 1
        dstsq = dot_product_table[0, 0]
        nvsd = 1
        barycentric_coordinates[0] = d1[0]
        search_direction[:] = simplex[0]
        iord[0] = 0
        # check line segment 1-2
        if not (d1[2] <= 0.0 or d2[2] <= 0.0):
            sum = d1[2] + d2[2]
            alsd[0] = d1[2] / sum
            alsd[1] = 1.0 - alsd[0]
            zsold[:] = simplex[1] + alsd[0] * (simplex[0] - simplex[1])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 1
        # check line segment 1-3
        if not (d1[4] <= 0.0 or d3[4] <= 0.0):
            sum = d1[4] + d3[4]
            alsd[0] = d1[4] / sum
            alsd[1] = 1.0 - alsd[0]
            zsold[:] = simplex[2] + alsd[0] * (simplex[0] - simplex[2])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 2
        # check face 1-2-3
        if not (d1[6] <= 0.0 or d2[6] <= 0.0 or d3[6] <= 0.0):
            sum = d1[6] + d2[6] + d3[6]
            alsd[0] = d1[6] / sum
            alsd[1] = d2[6] / sum
            alsd[2] = 1.0 - alsd[0] - alsd[1]
            zsold[:] = simplex[2] + alsd[0] * (simplex[0] - simplex[2]) + alsd[1] * (simplex[1] - simplex[2])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 3
                barycentric_coordinates[:] = alsd
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 1
                iord[2] = 2
        # check line segment 1-4
        if not (d1[8] <= 0.0 or d4[8] <= 0.0):
            sum = d1[8] + d4[8]
            alsd[0] = d1[8] / sum
            alsd[1] = 1.0 - alsd[0]
            zsold[:] = simplex[3] + alsd[0] * (simplex[0] - simplex[3])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 3
        # check face 1-2-4
        if not (d1[11] <= 0.0 or d2[11] <= 0.0 or d4[11] <= 0.0):
            sum = d1[11] + d2[11] + d4[11]
            alsd[0] = d1[11] / sum
            alsd[1] = d2[11] / sum
            alsd[2] = 1.0 - alsd[0] - alsd[1]
            zsold[:] = simplex[3] + alsd[0] * (simplex[0] - simplex[3]) + alsd[1] * (simplex[1] - simplex[3])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 3
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 1
                iord[2] = 3
        # check face 1-3-4
        if not (d1[12] <= 0.0 or d3[12] <= 0.0 or d4[12] <= 0.0):
            sum = d1[12] + d3[12] + d4[12]
            alsd[0] = d1[12] / sum
            alsd[2] = d3[12] / sum
            alsd[1] = 1.0 - alsd[0] - alsd[2]
            zsold[:] = simplex[3] + alsd[0] * (simplex[0] - simplex[3]) + alsd[2] * (simplex[2] - simplex[3])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 3
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 3
                iord[2] = 2
        # check the hull of all 4 points
        if not (d1[14] <= 0.0 or d2[14] <= 0.0 or d3[14] <= 0.0 or d4[14] <= 0.0):
            sum = d1[14] + d2[14] + d3[14] + d4[14]
            alsd[0] = d1[14] / sum
            alsd[1] = d2[14] / sum
            alsd[2] = d3[14] / sum
            alsd[3] = 1.0 - alsd[0] - alsd[1] - alsd[2]
            zsold[:] = alsd[0] * simplex[0] + alsd[1] * simplex[1] + alsd[2] * simplex[2] + alsd[3] * simplex[3]
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 4
                barycentric_coordinates[:] = alsd
                search_direction[:] = zsold
                iord[0] = 0
                iord[1] = 1
                iord[2] = 2
                iord[3] = 3
        # check vertex 2
        if dot_product_table[1, 1] < dstsq:
            nvsd = 1
            dstsq = dot_product_table[1, 1]
            barycentric_coordinates[0] = d2[1]
            search_direction[:] = simplex[1]
            iord[0] = 1
        # check vertex 3
        if dot_product_table[2, 2] < dstsq:
            nvsd = 1
            dstsq = dot_product_table[2, 2]
            barycentric_coordinates[0] = d3[3]
            search_direction[:] = simplex[2]
            iord[0] = 2
        # check vertex 4
        if dot_product_table[3, 3] < dstsq:
            nvsd = 1
            dstsq = dot_product_table[3, 3]
            barycentric_coordinates[0] = d4[7]
            search_direction[:] = simplex[3]
            iord[0] = 3
        # check line segment 2-3
        if not (d2[5] <= 0.0 or d3[5] <= 0.0):
            sum = d2[5] + d3[5]
            alsd[1] = d2[5] / sum
            alsd[0] = 1.0 - alsd[1]
            zsold[:] = simplex[2] + alsd[1] * (simplex[1] - simplex[2])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 2
                iord[1] = 1
        # check line segment 2-4
        if not (d2[9] <= 0.0 or d4[9] <= 0.0):
            sum = d2[9] + d4[9]
            alsd[1] = d2[9] / sum
            alsd[0] = 1.0 - alsd[1]
            zsold[:] = simplex[3] + alsd[1] * (simplex[1] - simplex[3])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 3
                iord[1] = 1
        # check line segment 3-4
        if not (d3[10] <= 0.0 or d4[10] <= 0.0):
            sum = d3[10] + d4[10]
            alsd[0] = d3[10] / sum
            alsd[1] = 1.0 - alsd[0]
            zsold[:] = simplex[3] + alsd[0] * (simplex[2] - simplex[3])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 2
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 2
                iord[1] = 3
        # check face 2-3-4
        if not (d2[13] <= 0.0 or d3[13] <= 0.0 or d4[13] <= 0.0):
            sum = d2[13] + d3[13] + d4[13]
            alsd[1] = d2[13] / sum
            alsd[2] = d3[13] / sum
            alsd[0] = 1.0 - alsd[1] - alsd[2]
            zsold[:] = simplex[3] + alsd[1] * (simplex[1] - simplex[3]) + alsd[2] * (simplex[2] - simplex[3])
            dstsqd = np.dot(zsold, zsold)
            if dstsqd < dstsq:
                dstsq = dstsqd
                nvsd = 3
                barycentric_coordinates[:nvsd] = alsd[:nvsd]
                search_direction[:] = zsold
                iord[0] = 3
                iord[1] = 1
                iord[2] = 2
    # final reordering
    risd = np.empty(4, dtype=int)
    risd[:n_simplex_points] = old_indices_polytope1[:n_simplex_points]
    rjsd = np.empty(4, dtype=int)
    rjsd[:n_simplex_points] = old_indices_polytope2[:n_simplex_points]
    yd = np.empty((4, 3), dtype=float)
    yd[:n_simplex_points] = simplex[:n_simplex_points]
    delld = np.empty((4, 4), dtype=float)
    for k in range(n_simplex_points):
        delld[k, :k + 1] = dot_product_table[k, :k + 1]

    n_simplex_points = nvsd
    for k in range(n_simplex_points):
        kk = iord[k]
        old_indices_polytope1[k] = risd[kk]
        old_indices_polytope2[k] = rjsd[kk]
        simplex[k] = yd[kk]
        for l in range(k):
            ll = iord[l]
            if kk >= ll:
                dot_product_table[k, l] = delld[kk, ll]
            else:
                dot_product_table[k, l] = delld[ll, kk]
        dot_product_table[k, k] = delld[kk, kk]
    backup = 1
    return dstsq, n_simplex_points, backup


def _revert_to_old_simplex(
        dot_product_table, indices_polytope1, indices_polytope2,
        old_dot_product_table, old_indices_polytope1, old_indices_polytope2,
        old_simplex, n_old_simplex_points, simplex):
    simplex[:n_old_simplex_points] = old_simplex[:n_old_simplex_points]
    indices_polytope1[:n_old_simplex_points] = old_indices_polytope1[:n_old_simplex_points]
    indices_polytope2[:n_old_simplex_points] = old_indices_polytope2[:n_old_simplex_points]
    dot_product_table[:n_old_simplex_points] = old_dot_product_table[:n_old_simplex_points]
    return n_old_simplex_points


def _add_new_point(
        dot_product_table, indices_polytope1, indices_polytope2,
        n_simplex_points, new_index1, new_index2, simplex, new_simplex_point):
    # Move first point to last spot
    indices_polytope1[n_simplex_points] = indices_polytope1[0]
    indices_polytope2[n_simplex_points] = indices_polytope2[0]
    simplex[n_simplex_points] = simplex[0]
    dot_product_table[n_simplex_points, :n_simplex_points] = dot_product_table[:n_simplex_points, 0]
    dot_product_table[n_simplex_points, n_simplex_points] = dot_product_table[0, 0]
    # Put new point in first spot
    indices_polytope1[0] = new_index1
    indices_polytope2[0] = new_index2
    simplex[0] = new_simplex_point
    # Update dot product table
    n_simplex_points += 1
    dot_product_table[:n_simplex_points, 0] = np.dot(simplex[:n_simplex_points], simplex[0])
    return n_simplex_points


def _save_old_simplex(
        dot_product_table, indices_polytope1, indices_polytope2,
        n_simplex_points, old_dot_product_table, old_indices_polytope1,
        old_indices_polytope2, old_simplex, simplex):
    # Save old values of n_simplex_points, indices_polytope1,
    # indices_polytope2, simplex and dot_product_table
    oldnvs = n_simplex_points
    old_simplex[:n_simplex_points] = simplex[:n_simplex_points]
    old_indices_polytope1[:n_simplex_points] = indices_polytope1[:n_simplex_points]
    old_indices_polytope2[:n_simplex_points] = indices_polytope2[:n_simplex_points]
    for k in range(n_simplex_points):
        old_dot_product_table[k, :k + 1] = dot_product_table[k, :k + 1]
    return oldnvs


def _reorder_simplex(
        dot_product_table, indices_polytope1, indices_polytope2, iord,
        n_simplex_points, old_dot_product_table, old_indices_polytope1,
        old_indices_polytope2, old_simplex, simplex):
    # If n_simplex_points == 4, rearrange dot_product_table[1, 0],
    # dot_product_table[2, 1] and dot_product_table[3, 0] in non decreasing
    # order
    if n_simplex_points == 4:
        iord[:3] = 0, 1, 2
        if dot_product_table[2, 0] < dot_product_table[1, 0]:
            iord[1] = 2
            iord[2] = 1
        ii = iord[1]
        if dot_product_table[3, 0] < dot_product_table[ii, 0]:
            iord[3] = iord[2]
            iord[2] = iord[1]
            iord[1] = 3
        else:
            ii = iord[2]
            if dot_product_table[3, 0] < dot_product_table[ii, 0]:
                iord[3] = iord[2]
                iord[2] = 3
            else:
                iord[3] = 3
        # Reorder indices_polytope1, indices_polytope2 simplex and dot_product_table
        for k in range(1, n_simplex_points):
            kk = iord[k]
            indices_polytope1[k] = old_indices_polytope1[kk]
            indices_polytope2[k] = old_indices_polytope2[kk]
            simplex[k] = old_simplex[kk]
            for l in range(k):
                ll = iord[l]
                if kk >= ll:
                    dot_product_table[k, l] = old_dot_product_table[kk, ll]
                else:
                    dot_product_table[k, l] = old_dot_product_table[ll, kk]
            dot_product_table[k, k] = old_dot_product_table[kk, kk]


def minkowski_sum(vertices1, vertices2):
    """Minkowski sum of two sets of vertices.

    Parameters
    ----------
    vertices1 : array, shape (n_vertices1, 3)
        First set of vertices.

    vertices2 : array, shape (n_vertices2, 3)
        Second set of vertices.

    Returns
    -------
    ms : array, shape (n_vertices1 * n_vertices2, 3)
        Sums of all pairs of vertices from first and second set.
    """
    return np.array([v1 + v2 for v1 in vertices1 for v2 in vertices2])



"""Distance between line and box.

Implementation adapted from 3D Game Engine Design by David H. Eberly.

Geometric Tools, Inc.
http://www.geometrictools.com
Copyright (c) 1998-2006.  All Rights Reserved

The Wild Magic Version 4 Foundation Library source code is supplied
under the terms of the license agreement
(http://www.geometrictools.com/License/Wm4FoundationLicense.pdf)
and may not be copied or disclosed except in accordance with the terms
of that agreement.
"""
import math

import numpy as np


def _line_to_box(line_point, line_direction, box2origin, size, origin2box=None):
    box_half_size = 0.5 * size

    # compute coordinates of line in box coordinate system
    if origin2box is None:
        origin2box = np.linalg.inv(box2origin)
    point_in_box = origin2box[:3, 3] + origin2box[:3, :3].dot(line_point)
    direction_in_box = origin2box[:3, :3].dot(line_direction)

    # Apply reflections so that direction vector has nonnegative components.
    direction_sign = np.ones(3)
    for i in range(3):
        if direction_in_box[i] < 0.0:
            point_in_box[i] = -point_in_box[i]
            direction_in_box[i] = -direction_in_box[i]
            direction_sign[i] = -1.0

    if direction_in_box[0] > 0.0:
        if direction_in_box[1] > 0.0:
            if direction_in_box[2] > 0.0:  # (+,+,+)
                sqr_dist, line_parameter = _case_no_zeros(
                    point_in_box, direction_in_box, box_half_size)
            else:  # (+,+,0)
                sqr_dist, line_parameter = _case_0(
                    0, 1, 2, point_in_box, direction_in_box, box_half_size)
        else:
            if direction_in_box[2] > 0.0:  # (+,0,+)
                sqr_dist, line_parameter = _case_0(
                    0, 2, 1, point_in_box, direction_in_box, box_half_size)
            else:  # (+,0,0)
                sqr_dist, line_parameter = _case_00(
                    0, 1, 2, point_in_box, direction_in_box, box_half_size)
    else:
        if direction_in_box[1] > 0.0:
            if direction_in_box[2] > 0.0:  # (0,+,+)
                sqr_dist, line_parameter = _case_0(
                    1, 2, 0, point_in_box, direction_in_box, box_half_size)
            else:  # (0,+,0)
                sqr_dist, line_parameter = _case_00(
                    1, 0, 2, point_in_box, direction_in_box, box_half_size)
        else:
            if direction_in_box[2] > 0.0:  # (0,0,+)
                sqr_dist, line_parameter = _case_00(
                    2, 0, 1, point_in_box, direction_in_box, box_half_size)
            else:  # (0,0,0)
                sqr_dist = _case_000(point_in_box, box_half_size)
                line_parameter = 0.0

    # compute closest point on line
    contact_point_line = line_point + line_parameter * line_direction

    # compute closest point on box
    # undo the reflections applied previously
    contact_point_box = box2origin[:3, 3] + box2origin[:3, :3].dot(
        direction_sign * point_in_box)

    return math.sqrt(sqr_dist), contact_point_line, contact_point_box, line_parameter


def _case_no_zeros(point_in_box, direction_in_box, box_half_size):
    point_m_edge = point_in_box - box_half_size

    prod_dx_py = direction_in_box[0] * point_m_edge[1]
    prod_dy_px = direction_in_box[1] * point_m_edge[0]

    if prod_dy_px >= prod_dx_py:
        prod_dz_px = direction_in_box[2] * point_m_edge[0]
        prod_dx_pz = direction_in_box[0] * point_m_edge[2]
        if prod_dz_px >= prod_dx_pz:
            # line intersects x = e0
            sqr_dist, line_parameter = _box_face(
                0, 1, 2, point_in_box, direction_in_box, point_m_edge,
                box_half_size)
        else:
            # line intersects z = e2
            sqr_dist, line_parameter = _box_face(
                2, 0, 1, point_in_box, direction_in_box, point_m_edge,
                box_half_size)
    else:
        prod_dz_py = direction_in_box[2] * point_m_edge[1]
        prod_dy_pz = direction_in_box[1] * point_m_edge[2]
        if prod_dz_py >= prod_dy_pz:
            # line intersects y = e1
            sqr_dist, line_parameter = _box_face(
                1, 2, 0, point_in_box, direction_in_box, point_m_edge,
                box_half_size)
        else:
            # line intersects z = e2
            sqr_dist, line_parameter = _box_face(
                2, 0, 1, point_in_box, direction_in_box, point_m_edge,
                box_half_size)
    return sqr_dist, line_parameter


def _box_face(i0, i1, i2, point_in_box, direction_in_box, point_m_edge, box_half_size):
    sqr_dist = 0.0
    point_p_edge = np.zeros(3)

    point_p_edge[i1] = point_in_box[i1] + box_half_size[i1]
    point_p_edge[i2] = point_in_box[i2] + box_half_size[i2]
    if direction_in_box[i0] * point_p_edge[i1] >= direction_in_box[i1] * point_m_edge[i0]:
        if direction_in_box[i0] * point_p_edge[i2] >= direction_in_box[i2] * point_m_edge[i0]:
            # v[i1] >= -e[i1], v[i2] >= -e[i2] (distance = 0)
            point_in_box[i0] = box_half_size[i0]
            point_in_box[i1] -= direction_in_box[i1] * point_m_edge[i0] / direction_in_box[i0]
            point_in_box[i2] -= direction_in_box[i2] * point_m_edge[i0] / direction_in_box[i0]
            line_parameter = -point_m_edge[i0] / direction_in_box[i0]
        else:
            # v[i1] >= -e[i1], v[i2] < -e[i2]
            l_sqr = (direction_in_box[i0] * direction_in_box[i0]
                     + direction_in_box[i2] * direction_in_box[i2])
            tmp = (l_sqr * point_p_edge[i1]
                   - direction_in_box[i1] * (direction_in_box[i0] * point_m_edge[i0]
                                             + direction_in_box[i2] * point_p_edge[i2]))
            if tmp <= 2.0 * l_sqr * box_half_size[i1]:
                t = tmp / l_sqr
                l_sqr += direction_in_box[i1] * direction_in_box[i1]
                tmp = point_p_edge[i1] - t
                delta = (direction_in_box[i0] * point_m_edge[i0]
                         + direction_in_box[i1] * tmp
                         + direction_in_box[i2] * point_p_edge[i2])
                line_parameter = -delta / l_sqr
                sqr_dist += (point_m_edge[i0] * point_m_edge[i0]
                             + tmp * tmp
                             + point_p_edge[i2] * point_p_edge[i2]
                             + delta * line_parameter)

                point_in_box[i0] = box_half_size[i0]
                point_in_box[i1] = t - box_half_size[i1]
                point_in_box[i2] = -box_half_size[i2]
            else:
                l_sqr += direction_in_box[i1] * direction_in_box[i1]
                delta = (direction_in_box[i0] * point_m_edge[i0]
                         + direction_in_box[i1] * point_m_edge[i1]
                         + direction_in_box[i2] * point_p_edge[i2])
                line_parameter = -delta / l_sqr
                sqr_dist += (point_m_edge[i0] * point_m_edge[i0]
                             + point_m_edge[i1] * point_m_edge[i1]
                             + point_p_edge[i2] * point_p_edge[i2]
                             + delta * line_parameter)

                point_in_box[i0] = box_half_size[i0]
                point_in_box[i1] = box_half_size[i1]
                point_in_box[i2] = -box_half_size[i2]
    else:
        if direction_in_box[i0] * point_p_edge[i2] >= direction_in_box[i2] * point_m_edge[i0]:
            # v[i1] < -e[i1], v[i2] >= -e[i2]
            l_sqr = (direction_in_box[i0] * direction_in_box[i0]
                     + direction_in_box[i1] * direction_in_box[i1])
            tmp = (l_sqr * point_p_edge[i2]
                   - direction_in_box[i2] * (direction_in_box[i0] * point_m_edge[i0]
                                             + direction_in_box[i1] * point_p_edge[i1]))
            if tmp <= 2.0 * l_sqr * box_half_size[i2]:
                t = tmp / l_sqr
                l_sqr += direction_in_box[i2] * direction_in_box[i2]
                tmp = point_p_edge[i2] - t
                delta = (direction_in_box[i0] * point_m_edge[i0]
                         + direction_in_box[i1] * point_p_edge[i1]
                         + direction_in_box[i2] * tmp)
                line_parameter = -delta / l_sqr
                sqr_dist += (point_m_edge[i0] * point_m_edge[i0]
                             + point_p_edge[i1] * point_p_edge[i1]
                             + tmp * tmp + delta * line_parameter)

                point_in_box[i0] = box_half_size[i0]
                point_in_box[i1] = -box_half_size[i1]
                point_in_box[i2] = t - box_half_size[i2]
            else:
                l_sqr += direction_in_box[i2] * direction_in_box[i2]
                delta = (direction_in_box[i0] * point_m_edge[i0]
                         + direction_in_box[i1] * point_p_edge[i1]
                         + direction_in_box[i2] * point_m_edge[i2])
                line_parameter = -delta / l_sqr
                sqr_dist += (point_m_edge[i0] * point_m_edge[i0]
                             + point_p_edge[i1] * point_p_edge[i1]
                             + point_m_edge[i2] * point_m_edge[i2]
                             + delta * line_parameter)

                point_in_box[i0] = box_half_size[i0]
                point_in_box[i1] = -box_half_size[i1]
                point_in_box[i2] = box_half_size[i2]
        else:
            # v[i1] < -e[i1], v[i2] < -e[i2]
            l_sqr = (direction_in_box[i0] * direction_in_box[i0]
                     + direction_in_box[i2] * direction_in_box[i2])
            tmp = (l_sqr * point_p_edge[i1]
                   - direction_in_box[i1] * (direction_in_box[i0] * point_m_edge[i0]
                                             + direction_in_box[i2] * point_p_edge[i2]))
            if tmp >= 0.0:
                # v[i1]-edge is closest
                if tmp <= 2.0 * l_sqr * box_half_size[i1]:
                    t = tmp / l_sqr
                    l_sqr += direction_in_box[i1] * direction_in_box[i1]
                    tmp = point_p_edge[i1] - t
                    delta = (direction_in_box[i0] * point_m_edge[i0]
                             + direction_in_box[i1] * tmp
                             + direction_in_box[i2] * point_p_edge[i2])
                    line_parameter = -delta / l_sqr
                    sqr_dist += (point_m_edge[i0] * point_m_edge[i0]
                                 + tmp * tmp
                                 + point_p_edge[i2] * point_p_edge[i2]
                                 + delta * line_parameter)

                    point_in_box[i0] = box_half_size[i0]
                    point_in_box[i1] = t - box_half_size[i1]
                    point_in_box[i2] = -box_half_size[i2]
                else:
                    l_sqr += direction_in_box[i1] * direction_in_box[i1]
                    delta = (direction_in_box[i0] * point_m_edge[i0]
                             + direction_in_box[i1] * point_m_edge[i1]
                             + direction_in_box[i2] * point_p_edge[i2])
                    line_parameter = -delta / l_sqr
                    sqr_dist += (point_m_edge[i0] * point_m_edge[i0]
                                 + point_m_edge[i1] * point_m_edge[i1]
                                 + point_p_edge[i2] * point_p_edge[i2]
                                 + delta * line_parameter)

                    point_in_box[i0] = box_half_size[i0]
                    point_in_box[i1] = box_half_size[i1]
                    point_in_box[i2] = -box_half_size[i2]
            else:
                l_sqr = (direction_in_box[i0] * direction_in_box[i0]
                         + direction_in_box[i1] * direction_in_box[i1])
                tmp = (l_sqr * point_p_edge[i2]
                       - direction_in_box[i2] * (direction_in_box[i0] * point_m_edge[i0]
                                                 + direction_in_box[i1] * point_p_edge[i1]))
                if tmp >= 0.0:
                    # v[i2]-edge is closest
                    if tmp <= 2.0 * l_sqr * box_half_size[i2]:
                        t = tmp / l_sqr
                        l_sqr += direction_in_box[i2] * direction_in_box[i2]
                        tmp = point_p_edge[i2] - t
                        delta = (direction_in_box[i0] * point_m_edge[i0]
                                 + direction_in_box[i1] * point_p_edge[i1]
                                 + direction_in_box[i2] * tmp)
                        line_parameter = -delta / l_sqr
                        sqr_dist += (point_m_edge[i0] * point_m_edge[i0]
                                     + point_p_edge[i1] * point_p_edge[i1]
                                     + tmp * tmp + delta * line_parameter)

                        point_in_box[i0] = box_half_size[i0]
                        point_in_box[i1] = -box_half_size[i1]
                        point_in_box[i2] = t - box_half_size[i2]
                    else:
                        l_sqr += direction_in_box[i2] * direction_in_box[i2]
                        delta = (direction_in_box[i0] * point_m_edge[i0]
                                 + direction_in_box[i1] * point_p_edge[i1]
                                 + direction_in_box[i2] * point_m_edge[i2])
                        line_parameter = -delta / l_sqr
                        sqr_dist += (point_m_edge[i0] * point_m_edge[i0]
                                     + point_p_edge[i1] * point_p_edge[i1]
                                     + point_m_edge[i2] * point_m_edge[i2]
                                     + delta * line_parameter)

                        point_in_box[i0] = box_half_size[i0]
                        point_in_box[i1] = -box_half_size[i1]
                        point_in_box[i2] = box_half_size[i2]
                else:
                    # (v[i1],v[i2])-corner is closest
                    l_sqr += direction_in_box[i2] * direction_in_box[i2]
                    delta = (direction_in_box[i0] * point_m_edge[i0]
                             + direction_in_box[i1] * point_p_edge[i1]
                             + direction_in_box[i2] * point_p_edge[i2])
                    line_parameter = -delta / l_sqr
                    sqr_dist += (point_m_edge[i0] * point_m_edge[i0]
                                 + point_p_edge[i1] * point_p_edge[i1]
                                 + point_p_edge[i2] * point_p_edge[i2]
                                 + delta * line_parameter)

                    point_in_box[i0] = box_half_size[i0]
                    point_in_box[i1] = -box_half_size[i1]
                    point_in_box[i2] = -box_half_size[i2]

    return sqr_dist, line_parameter


def _case_0(i0, i1, i2, point_in_box, direction_in_box, box_half_size):
    sqr_dist = 0.0
    point_m_edge0 = point_in_box[i0] - box_half_size[i0]
    point_m_edge1 = point_in_box[i1] - box_half_size[i1]
    prod0 = direction_in_box[i1] * point_m_edge0
    prod1 = direction_in_box[i0] * point_m_edge1

    if prod0 >= prod1:
        # line intersects P[i0] = e[i0]
        point_in_box[i0] = box_half_size[i0]

        point_p_edge1 = point_in_box[i1] + box_half_size[i1]
        delta = prod0 - direction_in_box[i0] * point_p_edge1
        if delta >= 0.0:
            inv_l_sqr = 1.0 / (direction_in_box[i0] * direction_in_box[i0]
                               + direction_in_box[i1] * direction_in_box[i1])
            sqr_dist += delta * delta * inv_l_sqr
            point_in_box[i1] = -box_half_size[i1]
            line_parameter = -(direction_in_box[i0] * point_m_edge0
                               + direction_in_box[i1] * point_p_edge1) * inv_l_sqr
        else:
            inv = 1.0 / direction_in_box[i0]
            point_in_box[i1] -= prod0 * inv
            line_parameter = -point_m_edge0 * inv
    else:
        # line intersects P[i1] = e[i1]
        point_in_box[i1] = box_half_size[i1]

        point_p_edge0 = point_in_box[i0] + box_half_size[i0]
        delta = prod1 - direction_in_box[i1] * point_p_edge0
        if delta >= 0.0:
            inv_l_sqr = 1.0 / (direction_in_box[i0] * direction_in_box[i0]
                               + direction_in_box[i1] * direction_in_box[i1])
            sqr_dist += delta * delta * inv_l_sqr
            point_in_box[i0] = -box_half_size[i0]
            line_parameter = -(direction_in_box[i0] * point_p_edge0
                               + direction_in_box[i1] * point_m_edge1) * inv_l_sqr
        else:
            inv = 1.0 / direction_in_box[i1]
            point_in_box[i0] -= prod1 * inv
            line_parameter = -point_m_edge1 * inv

    if point_in_box[i2] < -box_half_size[i2]:
        delta = point_in_box[i2] + box_half_size[i2]
        sqr_dist += delta * delta
        point_in_box[i2] = -box_half_size[i2]
    elif point_in_box[i2] > box_half_size[i2]:
        delta = point_in_box[i2] - box_half_size[i2]
        sqr_dist += delta * delta
        point_in_box[i2] = box_half_size[i2]

    return sqr_dist, line_parameter


def _case_00(i0, i1, i2, point_in_box, direction_in_box, box_half_size):
    line_parameter = (box_half_size[i0] - point_in_box[i0]) / direction_in_box[i0]
    point_in_box[i0] = box_half_size[i0]
    box_half_size_i12 = box_half_size[[i1, i2]]
    new_point_in_box = np.clip(
        point_in_box[[i1, i2]], -box_half_size_i12, box_half_size_i12)
    deltas = point_in_box[[i1, i2]] - new_point_in_box
    point_in_box[[i1, i2]] = new_point_in_box
    return np.dot(deltas, deltas), line_parameter


def _case_000(point_in_box, box_half_size):
    new_point_in_box = np.clip(point_in_box, -box_half_size, box_half_size)
    deltas = point_in_box - new_point_in_box
    point_in_box[:] = new_point_in_box
    return np.dot(deltas, deltas)



import math
import numpy as np
from ..geometry import line_from_pluecker
from ._line import point_to_line
from ._plane import plane_intersects_plane


def point_to_disk(point, center, radius, normal):
    """Compute the shortest distance between point and disk.

    Implementation adapted from 3D Game Engine Design by David H. Eberly.

    Geometric Tools, Inc.
    http://www.geometrictools.com
    Copyright (c) 1998-2006.  All Rights Reserved

    The Wild Magic Version 4 Foundation Library source code is supplied
    under the terms of the license agreement
    (http://www.geometrictools.com/License/Wm4FoundationLicense.pdf)
    and may not be copied or disclosed except in accordance with the terms
    of that agreement.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    center : array, shape (3,)
        Center of the disk.

    radius : float
        Radius of the disk.

    normal : array, shape (3,)
        Normal to the plane in which the disk lies.

    Returns
    -------
    dist : float
        The shortest distance between point and disk.

    closest_point_disk : array, shape (3,)
        Closest point on the disk.
    """
    # signed distance from point to plane of disk
    diff = point - center
    dist_to_plane = diff.dot(normal)

    # projection of P - C onto plane is Q - C = P - C - dist_to_plane * normal
    diff_in_plane = diff - dist_to_plane * normal
    sqr_len = diff_in_plane.dot(diff_in_plane)

    length = math.sqrt(sqr_len)
    t = radius
    if length != 0.0:
        t /= length
    closest_point_disk = center + min(1.0, t) * diff_in_plane

    return np.linalg.norm(point - closest_point_disk), closest_point_disk


def disk_to_disk(center1, radius1, normal1, center2, radius2, normal2, epsilon=1e-8):
    """Compute the shortest distance between two disks.

    Parameters
    ----------
    center1 : array, shape (3,)
        Center of the disk.

    radius1 : float
        Radius of the disk.

    normal1 : array, shape (3,)
        Normal to the plane in which the disk lies.

    center2 : array, shape (3,)
        Center of the disk.

    radius2 : float
        Radius of the disk.

    normal2 : array, shape (3,)
        Normal to the plane in which the disk lies.

    epsilon : float, optional (default: 1e-8)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between two disks.

    closest_point_disk1 : array, shape (3,)
        Closest point on the disk 1.

    closest_point_disk2 : array, shape (3,)
        Closest point on the disk 2.
    """
    # (1) test intersection first (source: https://stackoverflow.com/a/67116330/915743)
    # Plücker coordinates of intersection line
    line_direction, line_moment = plane_intersects_plane(center1, normal1, center2, normal2)
    line_point, line_direction = line_from_pluecker(line_direction, line_moment)
    h1, p1 = point_to_line(center1, line_point, line_direction)
    h2, p2 = point_to_line(center2, line_point, line_direction)
    ell = np.linalg.norm(p2 - p1)
    h = h1 + h2
    if abs(h) > epsilon:
        t1 = h1 * ell / h
        closest_to_both_disks = p1 - line_direction * t1
        if np.linalg.norm(closest_to_both_disks - center1) < radius1 and np.linalg.norm(closest_to_both_disks - center2) < radius2:
            return 0.0, closest_to_both_disks, closest_to_both_disks
    elif ell <= radius1 + radius2:  # both centers are on the common line
        contact = 0.5 * (center1 + center2)
        return 0.0, contact, contact

    # (2) no contact: simple iterative procedure
    # better solution: https://www.sciencedirect.com/science/article/pii/S0307904X0200080X
    _, contact_point_disk2 = point_to_disk(center1, center2, radius2, normal2)
    prev_dist = np.linalg.norm(center2 - center1)
    for i in range(20):
        _, contact_point_disk1 = point_to_disk(
            contact_point_disk2, center1, radius1, normal1)
        _, contact_point_disk2 = point_to_disk(
            contact_point_disk1, center2, radius2, normal2)
        dist = np.linalg.norm(contact_point_disk2 - contact_point_disk1)
        if prev_dist - dist < epsilon:
            break
        prev_dist = dist
    return np.linalg.norm(contact_point_disk2 - contact_point_disk1), contact_point_disk1, contact_point_disk2



"""Expanding polytope algorithm (EPA) for collision resolution after GJK."""
import numpy as np
from .colliders import Convex
from .utils import norm_vector


EDGES_PER_FACE = 3


def epa_vertices(simplex, vertices1, vertices2, epsilon=1e-8):
    """Expanding Polytope Algorithm (EPA).

    Find minimum translation vector to resolve collision.

    Parameters
    ----------
    simplex : array-like, shape (4, 3)
        Simplex of Minkowski distances obtained by GJK.

    vertices1 : array, shape (n_vertices1, 3)
        Vertices of the first convex shape.

    vertices2 : array, shape (n_vertices2, 3)
        Vertices of the second convex shape.

    epsilon : float, optional (default: 1e-8)
        Floating point tolerance

    Returns
    -------
    mtv : array, shape (3,)
        Minimum translation vector to be added to the second set of vertices
        or subtracted from the first set of vertices to resolve the collision.
        The norm of this vector is the penetration depth and the direction is
        the contact normal.
    """
    return epa(simplex, Convex(vertices1), Convex(vertices2), epsilon=epsilon)[0]


def epa(simplex, collider1, collider2, max_iter=64, max_loose_edges=32, max_faces=64, epsilon=1e-8):
    """Expanding Polytope Algorithm (EPA).

    Find minimum translation vector to resolve collision.

    Based on Kevin Moran's implementation:
        https://github.com/kevinmoran/GJK/blob/b38d923d268629f30b44c3cf6d4f9974bbcdb0d3/GJK.h
    (MIT License or Unlicense)

    Parameters
    ----------
    simplex : array, shape (4, 3)
        Simplex of Minkowski distances obtained by GJK.

    collider1 : Collider
        Convex collider 1.

    collider2 : Collider
        Convex collider 2.

    max_iter : int, optional (default: 64)
        Maximum number of iterations.

    max_loose_edges : int, optional (default: 32)
        Maximum number of loose edges per iteration.

    max_faces : int, optional (default: 64)
        Maximum number of faces in polytope.

    epsilon : float, optional (default: 1e-8)
        Floating point tolerance.

    Returns
    -------
    mtv : array, shape (3,)
        Minimum translation vector to be added to the second set of vertices
        or subtracted from the first set of vertices to resolve the collision.
        The norm of this vector is the penetration depth and the direction is
        the contact normal.

    faces : array, shape (n_faces, 4, 3)
        Faces in Minkowski difference space, defined by 3 points and its
        normal.

    success : bool
        EPA converged before maximum number of iterations was reached.
    """
    # Array of faces, each with 3 vertices and a normal
    faces = np.zeros((max_faces, 4, 3))
    # Keep track of edges we need to fix after removing faces
    loose_edges = np.zeros((max_loose_edges, 2, 3))

    n_faces = _simplex_to_faces(simplex, faces)

    for iteration in range(max_iter):
        min_dist, closest_face = _face_closest_to_origin(faces, n_faces)

        # Search normal to face that is closest to origin
        search_direction = faces[closest_face, 3]
        _, new_vertex1 = collider1.support_function(search_direction)
        _, new_vertex2 = collider2.support_function(-search_direction)
        new_point = new_vertex1 - new_vertex2

        if np.dot(new_point, search_direction) - min_dist < epsilon:
            # Convergence: new point is not significantly further from origin.
            # Dot product vertex with normal to resolve collision along normal.
            mtv = faces[closest_face, 3] * np.dot(new_point, search_direction)
            return mtv, faces[:n_faces], True

        n_faces, n_loose_edges = _find_loose_edges(faces, loose_edges, n_faces, new_point, max_loose_edges, epsilon)
        n_faces = _extend_polytope(faces, loose_edges, n_faces, n_loose_edges, new_point, max_faces)

    # Return most recent closest point
    mtv = faces[closest_face, 3] * np.dot(faces[closest_face, 0], faces[closest_face, 3])
    return mtv, faces[:n_faces], False


def _simplex_to_faces(simplex, faces):
    """Initialize with final simplex from GJK."""
    # ABC
    faces[0, :3] = simplex[:3]
    faces[0, 3] = norm_vector(np.cross(simplex[1] - simplex[0], simplex[2] - simplex[0]))
    # ACD
    faces[1, :3] = simplex[np.array((0, 2, 3), dtype=int)]
    faces[1, 3] = norm_vector(np.cross(simplex[2] - simplex[0], simplex[3] - simplex[0]))
    # ADB
    faces[2, :3] = simplex[np.array((0, 3, 1), dtype=int)]
    faces[2, 3] = norm_vector(np.cross(simplex[3] - simplex[0], simplex[1] - simplex[0]))
    # BDC
    faces[3, :3] = simplex[np.array((1, 3, 2), dtype=int)]
    faces[3, 3] = norm_vector(np.cross(simplex[3] - simplex[1], simplex[2] - simplex[1]))
    return 4


def _face_closest_to_origin(faces, n_faces):
    """Find face that is closest to origin."""
    dists = np.sum(faces[:n_faces, 0] * faces[:n_faces, 3], axis=1)
    closest_face = np.argmin(dists)
    return dists[closest_face], closest_face


def _find_loose_edges(faces, loose_edges, n_faces, new_points, max_loose_edges, epsilon):
    """Find all triangles that are facing p and store loose edges."""
    n_loose_edges = 0
    i = 0
    while i < n_faces:
        triangle_i_faces_p = np.dot(faces[i, 3], new_points - faces[i, 0]) > epsilon
        if triangle_i_faces_p:  # Remove it
            # Add removed triangle's edges to loose edge list.
            # If it is already there, remove it (both triangles it belonged to are gone)
            for j in range(EDGES_PER_FACE):
                current_edge = np.vstack([faces[i, j],
                                          faces[i, (j + 1) % 3]])
                found_edge = False
                k = 0
                while k < n_loose_edges:  # Check if current edge is already in list
                    if (np.linalg.norm(loose_edges[k, 1] - current_edge[0]) < epsilon and
                            np.linalg.norm(loose_edges[k, 0] - current_edge[1]) < epsilon):
                        # Edge is already in the list, remove it
                        # THIS ASSUMES EDGE CAN ONLY BE SHARED BY 2 TRIANGLES (which should be true)
                        # THIS ALSO ASSUMES SHARED EDGE WILL BE REVERSED IN THE TRIANGLES (which
                        # should be true provided every triangle is wound CCW)
                        # Overwrite current edge with last edge in list
                        loose_edges[k] = loose_edges[n_loose_edges - 1]
                        n_loose_edges -= 1
                        found_edge = True
                        # Exit loop because edge can only be shared once
                        break
                    k += 1

                if not found_edge:  # Add current edge to list
                    assert n_loose_edges < max_loose_edges
                    if n_loose_edges >= max_loose_edges:
                        break
                    loose_edges[n_loose_edges] = current_edge
                    n_loose_edges += 1

            # Remove triangle i from list
            faces[i] = faces[n_faces - 1]
            n_faces -= 1
            i -= 1
        i += 1
    return n_faces, n_loose_edges


def _extend_polytope(faces, loose_edges, n_faces, n_loose_edges, new_points, max_faces):
    """Reconstruct polytope with p added."""
    for i in range(n_loose_edges):
        assert n_faces < max_faces
        if n_faces >= max_faces:
            break
        faces[n_faces, :2] = loose_edges[i]
        faces[n_faces, 2] = new_points
        faces[n_faces, 3] = norm_vector(
            np.cross(faces[n_faces, 0] - faces[n_faces, 1],
                     faces[n_faces, 0] - faces[n_faces, 2]))
        if np.linalg.norm(faces[n_faces, 3]) < 0.5:  # TODO is this the right solution?
            continue
        _fix_ccw_normal_direction(faces, n_faces)
        n_faces += 1
    return n_faces


def _fix_ccw_normal_direction(faces, face_idx, bias=1e-6):
    """Correct wrong normal direction to maintain CCW winding."""
    # Use bias in case dot result is only slightly < 0 (because origin is on face)
    if np.dot(faces[face_idx, 0], faces[face_idx, 3]) + bias < 0.0:
        temp = faces[face_idx, 0]
        faces[face_idx, 0] = faces[face_idx, 1]
        faces[face_idx, 1] = temp
        faces[face_idx, 3] = -faces[face_idx, 3]



import numpy as np
from ..geometry import (
    convert_rectangle_to_segment, convert_segment_to_line)
from ._line import _line_to_line_segment
import pytransform3d.rotations as pr


def point_to_rectangle(point, rectangle_center, rectangle_axes,
                       rectangle_lengths):
    """Compute the shortest distance from point to rectangle.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    rectangle_center : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle.

    Returns
    -------
    dist : float
        The shortest distance between the point and the rectangle.

    closest_point_rectangle : array, shape (3,)
        Closest point on the rectangle.
    """
    rectangle_coordinates = rectangle_axes.dot(point - rectangle_center)

    rectangle_half_lengths = 0.5 * rectangle_lengths
    rectangle_coordinates = np.clip(
        rectangle_coordinates, -rectangle_half_lengths, rectangle_half_lengths)

    closest_point_rectangle = (
        rectangle_center + rectangle_coordinates.dot(rectangle_axes))

    return (np.linalg.norm(point - closest_point_rectangle),
            closest_point_rectangle)


def line_to_rectangle(
        line_point, line_direction,
        rectangle_center, rectangle_axes, rectangle_lengths, epsilon=1e-6):
    """Compute the shortest distance between line and rectangle.

    Implementation adapted from 3D Game Engine Design by David H. Eberly.

    Geometric Tools, Inc.
    http://www.geometrictools.com
    Copyright (c) 1998-2006.  All Rights Reserved

    The Wild Magic Version 4 Foundation Library source code is supplied
    under the terms of the license agreement
    (http://www.geometrictools.com/License/Wm4FoundationLicense.pdf)
    and may not be copied or disclosed except in accordance with the terms
    of that agreement.

    Parameters
    ----------
    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.

    rectangle_center : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between line and rectangle.

    contact_point_line : array, shape (3,)
        Closest point on the line.

    contact_point_rectangle : array, shape (3,)
        Closest point on the rectangle.
    """
    return _line_to_rectangle(
        line_point, line_direction,
        rectangle_center, rectangle_axes, rectangle_lengths, epsilon)[:3]


def _line_to_rectangle(
        line_point, line_direction,
        rectangle_center, rectangle_axes, rectangle_lengths, epsilon=1e-6):
    rectangle_half_lengths = 0.5 * rectangle_lengths

    # Test if line intersects rectangle. If so, the squared distance is zero.
    intersects, result = _line_intersects_rectangle(
        line_point, line_direction, rectangle_center,
        rectangle_axes, rectangle_half_lengths, epsilon)
    if intersects:
        return result

    rectangle_extents = rectangle_half_lengths[:, np.newaxis] * rectangle_axes
    # Either (1) the line is not parallel to the rectangle and the point of
    # intersection of the line and the plane of the rectangle is outside the
    # rectangle or (2) the line and rectangle are parallel. Regardless, the
    # closest point on the rectangle is on an edge of the rectangle. Compare
    # the line to all four edges of the rectangle.
    best_dist = np.finfo(float).max
    for i1 in range(2):
        for i0 in range(2):
            segment_end, segment_start = convert_rectangle_to_segment(
                rectangle_center, rectangle_extents, i0, i1)
            dist, contact_point_line, contact_point_segment, l_closest, s_closest = _line_to_line_segment(
                line_point, line_direction, segment_start, segment_end, epsilon=epsilon)

            if dist < best_dist:
                best_contact_point_line = contact_point_line
                best_contact_point_rectangle = contact_point_segment
                best_dist = dist
                best_line_parameter = l_closest
            if best_dist < epsilon:
                break

    return best_dist, best_contact_point_line, best_contact_point_rectangle, best_line_parameter


def _line_intersects_rectangle(
        line_point, line_direction, rectangle_center, rectangle_axes,
        rectangle_half_lengths, epsilon):
    rectangle_normal = np.cross(rectangle_axes[0], rectangle_axes[1])
    if abs(rectangle_normal.dot(line_direction)) > epsilon:
        # The line and rectangle are not parallel, so the line intersects
        # the plane of the rectangle.
        diff = line_point - rectangle_center
        u, v = pr.plane_basis_from_normal(line_direction)
        udd = rectangle_axes.dot(u)
        vdd = rectangle_axes.dot(v)
        uddiff = u.dot(diff)
        vddiff = v.dot(diff)
        det = udd[0] * vdd[1] - udd[1] * vdd[0]

        # Rectangle coordinates for the point of intersection.
        s = np.array([(vdd[1] * uddiff - udd[1] * vddiff),
                      (udd[0] * vddiff - vdd[0] * uddiff)]) / det

        if abs(s[0]) <= rectangle_half_lengths[0] and abs(s[1]) <= rectangle_half_lengths[1]:
            # Line parameter for the point of intersection.
            line_direction_d_d = rectangle_axes.dot(line_direction)
            line_direction_dot_diff = line_direction.dot(diff)
            line_parameter = np.dot(s, line_direction_d_d) - line_direction_dot_diff

            # The intersection point is inside or on the rectangle.
            contact_point_line = line_point + line_parameter * line_direction
            contact_point_rectangle = rectangle_center + s.dot(rectangle_axes)
            return True, (0.0, contact_point_line, contact_point_rectangle, line_parameter)
    return False, None


def line_segment_to_rectangle(
        segment_start, segment_end,
        rectangle_center, rectangle_axes, rectangle_lengths, epsilon=1e-6):
    """Compute the shortest distance between line segment and rectangle.

    Parameters
    ----------
    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    rectangle_center : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between line segment and rectangle.

    closest_point_line_segment : array, shape (3,)
        Closest point on the line segment.

    closest_point_rectangle : array, shape (3,)
        Closest point on the rectangle.
    """
    segment_direction, segment_length = convert_segment_to_line(
        segment_start, segment_end)

    distance, closest_point_line_segment, closest_point_rectangle, t = \
        _line_to_rectangle(
            segment_start, segment_direction,
            rectangle_center, rectangle_axes, rectangle_lengths, epsilon)

    if t < 0:
        distance, closest_point_rectangle = point_to_rectangle(
            segment_start, rectangle_center, rectangle_axes, rectangle_lengths)
        closest_point_line_segment = segment_start
    elif t > segment_length:
        distance, closest_point_rectangle = point_to_rectangle(
            segment_end, rectangle_center, rectangle_axes, rectangle_lengths)
        closest_point_line_segment = segment_end

    return distance, closest_point_line_segment, closest_point_rectangle


def rectangle_to_rectangle(
        rectangle_center1, rectangle_axes1, rectangle_lengths1,
        rectangle_center2, rectangle_axes2, rectangle_lengths2, epsilon=1e-6):
    """Compute the shortest distance between two rectangles.

    Implementation adapted from 3D Game Engine Design by David H. Eberly.

    Geometric Tools, Inc.
    http://www.geometrictools.com
    Copyright (c) 1998-2006.  All Rights Reserved

    The Wild Magic Version 4 Foundation Library source code is supplied
    under the terms of the license agreement
    (http://www.geometrictools.com/License/Wm4FoundationLicense.pdf)
    and may not be copied or disclosed except in accordance with the terms
    of that agreement.

    Parameters
    ----------
    rectangle_center1 : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes1 : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths1 : array, shape (2,)
        Lengths of the two sides of the rectangle.

    rectangle_center2 : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes2 : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths2 : array, shape (2,)
        Lengths of the two sides of the rectangle.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between two rectangles.

    closest_point_line_segment : array, shape (3,)
        Closest point on the line segment.

    closest_point_rectangle : array, shape (3,)
        Closest point on the rectangle.
    """
    # compare edges of rectangle0 to the interior of rectangle1
    best_dist = np.finfo(float).max

    rectangle_half_lengths1 = 0.5 * rectangle_lengths1
    rectangle_extents1 = rectangle_half_lengths1[:, np.newaxis] * rectangle_axes1
    for i1 in range(2):
        for i0 in range(2):
            segment_end, segment_start = convert_rectangle_to_segment(
                rectangle_center1, rectangle_extents1, i0, i1)

            dist, contact_point_rectangle1, contact_point_rectangle2 = line_segment_to_rectangle(
                segment_start, segment_end, rectangle_center2, rectangle_axes2, rectangle_lengths2)

            if dist < best_dist:
                best_contact_point_rectangle1 = contact_point_rectangle1
                best_contact_point_rectangle2 = contact_point_rectangle2
                best_dist = dist
            if dist <= epsilon:
                break

    # compare edges of rectangle1 to the interior of rectangle0
    rectangle_half_lengths2 = 0.5 * rectangle_lengths2
    rectangle_extents2 = rectangle_half_lengths2[:, np.newaxis] * rectangle_axes2
    for i1 in range(2):
        for i0 in range(2):
            segment_end, segment_start = convert_rectangle_to_segment(
                rectangle_center2, rectangle_extents2, i0, i1)

            dist, contact_point_rectangle2, contact_point_rectangle1 = line_segment_to_rectangle(
                segment_start, segment_end, rectangle_center1, rectangle_axes1, rectangle_lengths1)

            if dist < best_dist:
                best_contact_point_rectangle1 = contact_point_rectangle1
                best_contact_point_rectangle2 = contact_point_rectangle2
                best_dist = dist
            if dist <= epsilon:
                break

    return best_dist, best_contact_point_rectangle1, best_contact_point_rectangle2



"""Containment methods compute bounding volumes."""
import numpy as np
from .geometry import convert_box_to_vertices


def axis_aligned_bounding_box(P):
    """Compute axis-aligned bounding box (AABB) that contains points.

    Parameters
    ----------
    P : array, shape (n_points, 3)
        3D points.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    return np.min(P, axis=0), np.max(P, axis=0)


def sphere_aabb(center, radius):
    """Compute axis-aligned bounding box of sphere.

    Parameters
    ----------
    center : array, shape (3,)
        Center of the sphere.

    radius : float
        Radius of the sphere.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    return center - radius, center + radius


def box_aabb(box2origin, size):
    """Compute axis-aligned bounding box of an oriented box.

    Parameters
    ----------
    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Size of the box.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    vertices = convert_box_to_vertices(box2origin, size)
    return axis_aligned_bounding_box(vertices)


def cylinder_aabb(cylinder2origin, radius, length):
    """Compute axis-aligned bounding box of cylinder.

    Parameters
    ----------
    cylinder2origin : array, shape (4, 4)
        Pose of the cylinder.

    radius : float
        Radius of the cylinder.

    length : float
        Length of the cylinder.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    # AABB of a cylinder is the same as the AABB of its caps,
    # see https://iquilezles.org/articles/diskbbox/
    axis = cylinder2origin[:3, 2]
    extent = 0.5 * length * np.abs(axis) + radius * np.sqrt(1.0 - axis * axis)
    return cylinder2origin[:3, 3] - extent, cylinder2origin[:3, 3] + extent


def capsule_aabb(capsule2origin, radius, height):
    """Compute axis-aligned bounding box of a capsule.

    Parameters
    ----------
    capsule2origin : array, shape (4, 4)
        Pose of the capsule.

    radius : float
        Radius of the capsule.

    height : float
        Height of the capsule.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    extent = 0.5 * height * np.abs(capsule2origin[:3, 2]) + radius
    return capsule2origin[:3, 3] - extent, capsule2origin[:3, 3] + extent



"""Colliders used for collision detection with GJK algorithm."""
import abc
import warnings
import numpy as np
from pytransform3d import urdf
import pytransform3d.visualizer as pv
from .geometry import (
    capsule_extreme_along_direction, cylinder_extreme_along_direction,
    convert_box_to_vertices)
from .containment import (
    axis_aligned_bounding_box, sphere_aabb, box_aabb, cylinder_aabb,
    capsule_aabb)
from .urdf_utils import self_collision_whitelists
from aabbtree import AABB, AABBTree


class BoundingVolumeHierarchy:
    """Bounding volume hierarchy (BVH) for broad phase collision detection.

    Wraps multiple colliders that are connected through transformations.
    In addition, these colliders are stored in an AABB tree for broad phase
    collision detection.

    Parameters
    ----------
    tm : pytransform3d.transform_manager.TransformManager
        Transform manager that stores the transformations.

    base_frame : str
        Name of the base frame in which colliders are represented.

    Attributes
    ----------
    aabbtree_ : AABBTree
        Tree of axis-aligned bounding boxes.

    colliders_ : dict
        Maps frames of collision objects to colliders.

    self_collision_whitelists_ : dict
        Whitelists for self-collision detection in case this BVH represents
        a robot.
    """
    def __init__(self, tm, base_frame):
        self.tm = tm
        self.base_frame = base_frame
        self.collider_frames = set()
        self.aabbtree_ = AABBTree()
        self.colliders_ = {}
        self.self_collision_whitelists_ = {}

    def fill_tree_with_colliders(
            self, tm, make_artists=False,
            fill_self_collision_whitelists=False):
        """Fill tree with colliders from URDF transform manager.

        Parameters
        ----------
        tm : pytransform3d.urdf.UrdfTransformManager
            Transform manager that has colliders.

        make_artists : bool, optional (default: False)
            Create artist for visualization for each collision object.

        fill_self_collision_whitelists : bool, optional (default: False)
            Fill whitelists for self collision detection. All collision
            objects connected to the current link, child, and parent links
            will be ignored.
        """
        for obj in tm.collision_objects:
            try:
                collider = self._make_collider(tm, obj, make_artists)
                self.add_collider(obj.frame, collider)
            except RuntimeError as e:
                warnings.warn(str(e))

        if fill_self_collision_whitelists:
            self.self_collision_whitelists_ = self_collision_whitelists(tm)

    def _make_collider(self, tm, obj, make_artists):
        A2B = tm.get_transform(obj.frame, self.base_frame)
        if isinstance(obj, urdf.Sphere):
            collider = Sphere(center=A2B[:3, 3], radius=obj.radius)
        elif isinstance(obj, urdf.Box):
            collider = Box(A2B, obj.size)
        elif isinstance(obj, urdf.Cylinder):
            collider = Cylinder(
                cylinder2origin=A2B, radius=obj.radius,
                length=obj.length)
        else:
            assert isinstance(obj, urdf.Mesh)
            collider = Mesh(obj.filename, A2B, obj.scale)
        if make_artists:
            collider.make_artist()
        return collider

    def add_collider(self, frame, collider):
        """Add collider.

        Parameters
        ----------
        frame : str
            Frame in which the collider is located.

        collider : ConvexCollider
            Collider.
        """
        self.collider_frames.add(frame)
        self.colliders_[frame] = collider
        self.aabbtree_.add(collider.aabb(), (frame, collider))

    def update_collider_poses(self):
        """Update poses of all colliders from transform manager."""
        self.aabbtree_ = AABBTree()
        for frame in self.colliders_:
            A2B = self.tm.get_transform(frame, self.base_frame)
            collider = self.colliders_[frame]
            collider.update_pose(A2B)
            self.aabbtree_.add(collider.aabb(), (frame, collider))

    def get_colliders(self):
        """Get all colliders.

        Returns
        -------
        colliders : list
            List of colliders.
        """
        return self.colliders_.values()

    def get_artists(self):
        """Get all artists.

        Returns
        -------
        artists : list
            List of artists.
        """
        return [collider.artist_ for collider in self.colliders_.values()
                if collider.artist_ is not None]

    def aabb_overlapping_colliders(self, collider, whitelist=()):
        """Get colliders with an overlapping AABB.

        This function performs broad phase collision detection with a bounding
        volume hierarchy, where the bounding volumes are axis-aligned bounding
        boxes.

        Parameters
        ----------
        collider : ConvexCollider
            Collider.

        whitelist : sequence
            Names of frames to which collisions are allowed.

        Returns
        -------
        colliders : dict
            Maps frame names to colliders with overlapping AABB.
        """
        aabb = collider.aabb()
        colliders = dict(self.aabbtree_.overlap_values(aabb))
        for frame in whitelist:
            colliders.pop(frame, None)
        return colliders

    def get_collider_frames(self):
        """Get collider frames.

        Returns
        -------
        collider_frames : set
            Collider frames.
        """
        return self.collider_frames


# for backwards compatibility:
ColliderTree = BoundingVolumeHierarchy


class ConvexCollider(abc.ABC):
    """Convex collider base class.

    Parameters
    ----------
    vertices : iterable
        Vertices of the convex collider.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.

    Attributes
    ----------
    vertices_ : iterable
        Vertices of the convex collider.

    artist_ : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, vertices, artist=None):
        self.vertices_ = vertices
        self.artist_ = artist

    @abc.abstractmethod
    def make_artist(self, c=None):
        """Make artist that represents this collider.

        Parameters
        ----------
        c : array-like, shape (3,), optional (default: None)
            Color of artist.
        """

    @abc.abstractmethod
    def first_vertex(self):
        """Get any vertex from collider to initialize GJK algorithm.

        Returns
        -------
        vertex : array, shape (3,)
            Vertex from collider.
        """

    @abc.abstractmethod
    def support_function(self, search_direction):
        """Support function for collider.

        Parameters
        ----------
        search_direction : array, shape (3,)
            Direction in which we search for extreme point of the collider.

        Returns
        -------
        extreme_point : array, shape (3,)
            Extreme point along search direction.
        """

    def compute_point(self, barycentric_coordinates, indices):
        """Compute point from barycentric coordinates.

        Parameters
        ----------
        barycentric_coordinates : array, shape (n_vertices,)
            Barycentric coordinates of the point that we compute.

        indices : array, shape (n_vertices,)
            Vertex indices to which the barycentric coordinates apply.

        Returns
        -------
        point : array, shape (3,)
            Point that we compute from barycentric coordinates.
        """
        return np.dot(barycentric_coordinates,
                      np.array([self.vertices_[i] for i in indices]))

    @abc.abstractmethod
    def update_pose(self, pose):
        """Update pose of collider.

        Parameters
        ----------
        pose : array, shape (4, 4)
            New pose of the collider.
        """

    @abc.abstractmethod
    def aabb(self):
        """Get axis-aligned bounding box.

        Returns
        -------
        aabb : AABB
            Axis-aligned bounding box.
        """


class Convex(ConvexCollider):
    """Wraps convex hull of a set of vertices for GJK algorithm.

    Parameters
    ----------
    vertices : array, shape (n_vertices, 3)
        Vertices of the convex shape.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Artist for visualizer.
    """
    def __init__(self, vertices, artist=None):
        super(Convex, self).__init__(vertices, artist)

    def make_artist(self, c=None):
        self.artist_ = pv.PointCollection3D(self.vertices_, s=0.005, c=c)

    def first_vertex(self):
        return self.vertices_[0]

    def support_function(self, search_direction):
        idx = np.argmax(self.vertices_.dot(search_direction))
        return idx, self.vertices_[idx]

    def compute_point(self, barycentric_coordinates, indices):
        return np.dot(barycentric_coordinates, self.vertices_[indices])

    def update_pose(self, vertices):
        self.vertices_ = vertices
        if self.artist_ is not None:
            self.artist_.set_data(self.vertices_)

    def aabb(self):
        mins, maxs = axis_aligned_bounding_box(self.vertices_)
        return AABB(np.array([mins, maxs]).T)


class Box(Convex):
    """Wraps box for GJK algorithm.

    Parameters
    ----------
    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Sizes of the box along its axes.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, box2origin, size, artist=None):
        super(Box, self).__init__(
            convert_box_to_vertices(box2origin, size), artist)
        self.box2origin = box2origin
        self.size = size

    def make_artist(self, c=None):
        self.artist_ = pv.Box(size=self.size, A2B=self.box2origin, c=c)

    def update_pose(self, pose):
        self.box2origin = pose
        self.vertices_ = convert_box_to_vertices(pose, self.size)
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = box_aabb(self.box2origin, self.size)
        return AABB(np.array([mins, maxs]).T)


class Mesh(Convex):
    """Wraps mesh for GJK algorithm (we assume a convex mesh).

    Parameters
    ----------
    filename : str
        Path to mesh file.

    A2B : array, shape (4, 4)
        Center of the mesh.

    scale : float, optional (default: 1)
        Scaling of the mesh.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, filename, A2B, scale=1.0, artist=None):
        import pytransform3d.visualizer as pv
        if artist is None:
            artist = pv.Mesh(filename=filename, A2B=A2B, s=scale)
        vertices = np.asarray(artist.mesh.vertices)
        super(Mesh, self).__init__(vertices, artist)

    def make_artist(self, c=None):
        assert self.artist_ is not None

    def update_pose(self, pose):
        self.artist_.set_data(pose)
        self.vertices_ = np.asarray(self.artist_.mesh.vertices)


class Cylinder(ConvexCollider):
    """Wraps cylinder for GJK algorithm.

    Parameters
    ----------
    cylinder2origin : array, shape (4, 4)
        Pose of the cylinder.

    radius : float
        Radius of the cylinder.

    length : float
        Length of the cylinder.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, cylinder2origin, radius, length, artist=None):
        super(Cylinder, self).__init__([], artist)
        self.cylinder2origin = cylinder2origin
        self.radius = radius
        self.length = length

    def make_artist(self, c=None):
        self.artist_ = pv.Cylinder(
            length=self.length, radius=self.radius, A2B=self.cylinder2origin,
            c=c)

    def first_vertex(self):
        vertex = self.cylinder2origin[:3, 3] + 0.5 * self.length * self.cylinder2origin[:3, 2]
        self.vertices_.append(vertex)
        return vertex

    def support_function(self, search_direction):
        vertex = cylinder_extreme_along_direction(
            search_direction, self.cylinder2origin, self.radius, self.length)
        vertex_idx = len(self.vertices_)
        self.vertices_.append(vertex)
        return vertex_idx, vertex

    def update_pose(self, pose):
        self.cylinder2origin = pose
        self.vertices_ = []
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = cylinder_aabb(
            self.cylinder2origin, self.radius, self.length)
        return AABB(np.array([mins, maxs]).T)


class Capsule(ConvexCollider):
    """Wraps capsule for GJK algorithm.

    Parameters
    ----------
    capsule2origin : array, shape (4, 4)
        Pose of the capsule.

    radius : float
        Radius of the capsule.

    height : float
        Height of the capsule.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, capsule2origin, radius, height, artist=None):
        super(Capsule, self).__init__([], artist)
        self.capsule2origin = capsule2origin
        self.radius = radius
        self.height = height

    def make_artist(self, c=None):
        self.artist_ = pv.Capsule(
            height=self.height, radius=self.radius, A2B=self.capsule2origin,
            c=c)

    def first_vertex(self):
        vertex = self.capsule2origin[:3, 3] - (self.radius + 0.5 * self.height) * self.capsule2origin[:3, 2]
        self.vertices_.append(vertex)
        return vertex

    def support_function(self, search_direction):
        vertex = capsule_extreme_along_direction(
            search_direction, self.capsule2origin, self.radius, self.height)
        vertex_idx = len(self.vertices_)
        self.vertices_.append(vertex)
        return vertex_idx, vertex

    def update_pose(self, pose):
        self.capsule2origin = pose
        self.vertices_ = []
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = capsule_aabb(
            self.capsule2origin, self.radius, self.height)
        return AABB(np.array([mins, maxs]).T)


class Sphere(ConvexCollider):
    """Wraps sphere for GJK algorithm.

    Parameters
    ----------
    center : array, shape (3,)
        Center of the sphere.

    radius : float
        Radius of the sphere.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, center, radius, artist=None):
        super(Sphere, self).__init__([], artist)
        self.c = center
        self.radius = radius

    def make_artist(self, c=None):
        sphere2origin = np.eye(4)
        sphere2origin[:3, 3] = self.c
        self.artist_ = pv.Sphere(radius=self.radius, A2B=sphere2origin, c=c)

    def first_vertex(self):
        vertex = self.c + np.array([0, 0, self.radius])
        self.vertices_.append(vertex)
        return vertex

    def support_function(self, search_direction):
        # Similar implementation:
        # https://github.com/kevinmoran/GJK/blob/b38d923d268629f30b44c3cf6d4f9974bbcdb0d3/Collider.h#L33
        # (Copyright (c) 2017 Kevin Moran, MIT License or Unlicense)
        s_norm = np.linalg.norm(search_direction)
        if s_norm == 0.0:
            vertex = self.c + np.array([0, 0, self.radius])
        else:
            vertex = self.c + search_direction / s_norm * self.radius
        vertex_idx = len(self.vertices_)
        self.vertices_.append(vertex)
        return vertex_idx, vertex

    def update_pose(self, pose):
        self.c = pose[:3, 3]
        self.vertices_ = []
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = sphere_aabb(self.c, self.radius)
        return AABB(np.array([mins, maxs]).T)
